"""The purpose of this script is to estimate parameters of the export supply function

- marginal propensity to export sigma_{ik}

- constant tetha_{ik}

X_ik = tetha_{ik} + sigma_{ik} I_{wkt}

Because world import I_{wkt} is not equal to world export X_{wkt},
the estimation is performed with world exports instead.

- See Buongiorno 2022 "GFPMX: A Cobweb Model of the Global Forest Sector, with an
  Application to the Impact of the COVID-19 Pandemic"

"""

import pandas
import xarray
import numpy as np
import seaborn
from sklearn.linear_model import LinearRegression
from matplotlib import pyplot as plt
from biotrade.faostat import faostat
from cobwood.gfpmx import GFPMX

# Load historical data
gfpmxb2021 = GFPMX(input_dir="gfpmx_base2021", base_year=2021, scenario="base_2021")


############################################################
# Estimate the industrial roundwood export supply function #
############################################################
def estimate_export_supply(ds, variable, country):
    """Estimate the export supply function
    modifies the dataset in place by adding new coefficient and constant"""
    X = ds[variable].loc["WORLD"].values.reshape(-1, 1)
    Y = ds[variable].loc[country].values.reshape(-1, 1)
    # TODO: display p values and errors of the estimates and R square
    model = LinearRegression()
    model.fit(X, Y)
    print(ds.product, country, model.intercept_[0], model.coef_[0, 0])
    # Store coefficients
    ds["exp_constant_2"].loc[country] = model.intercept_[0]
    ds["exp_mpte_2"].loc[country] = model.coef_[0, 0]
    # Store predicted values
    ds["exp_predicted"].loc[country] = model.predict(X).squeeze()


def prepare_dataset_for_estimation(ds):
    """Prepare the dataset for estimation
    returns a new new dataset"""
    # Keep only data related for the historical period
    ds_out = ds.loc[{"year": ds.coords["year"] <= 2021}]
    # Initialize empty new coefficients as 1D datasets
    ds_out["exp_mpte_2"] = xarray.full_like(ds_out["exp_constant"], fill_value=np.nan)
    ds_out["exp_constant_2"] = xarray.full_like(
        ds_out["exp_constant"], fill_value=np.nan
    )
    ds_out["exp_predicted"] = xarray.full_like(ds_out["exp"], fill_value=np.nan)
    return ds_out


# Prepare the datasets
irw = prepare_dataset_for_estimation(gfpmxb2021["indround"])
sw = prepare_dataset_for_estimation(gfpmxb2021["sawn"])

# Estimate the export supply functions
for this_country in irw["country"].loc[irw.c].values:
    estimate_export_supply(irw, "exp", this_country)
    estimate_export_supply(sw, "exp", this_country)

# Change Czechia historical export to the mean of last 5 years
# Re-estimate the indround model for that country
yr = irw.coords["year"]
cz_mean = irw["exp"].loc["Czechia", (2011 <= yr) & (yr <= 2015)].mean().values
# Initialize empty new coefficients as 1D datasets
irw["exp_2"] = irw["exp"].copy()
irw["exp_2"].loc["Czechia", (yr > 2015)] = cz_mean
estimate_export_supply(irw, "exp", "Czechia")
estimate_export_supply(irw, "exp_2", "Czechia")

# Change Czechia historical export to the mean of last 5 years
# Re-estimate the sawn model for that country
yr = sw.coords["year"]
cz_mean = sw["exp"].loc["Czechia", (2011 <= yr) & (yr <= 2015)].mean().values
# Initialize empty new coefficients as 1D datasets
sw["exp_2"] = sw["exp"].copy()
sw["exp_2"].loc["Czechia", (yr > 2015)] = cz_mean
estimate_export_supply(sw, "exp", "Czechia")
estimate_export_supply(sw, "exp_2", "Czechia")

# Czechia indround
# existing coefficients: 12971.611067935588 0.0364487321930629
# new coefficients: -395.63598204559366 0.030772674789748013
# Czechia sawn
# existing coefficients: 1593.445938593596 0.0136056235161746
# new coefficients: 1344.7487441529158 0.0025397644268203446
for ds in [irw, sw]:
    ds_cz = ds.loc[{"country": "Czechia"}]
    print(
        f"Czechia {ds.product}\nexisting coefficients:",
        ds_cz["exp_constant"].values,
        ds_cz["exp_marginal_propensity_to_export"].values,
        "\nnew coefficients:",
        ds_cz["exp_constant_2"].values,
        ds_cz["exp_mpte_2"].values,
    )

# Project to 2050 based on a rough projection of irw world imports
irw_world = pandas.DataFrame(
    {
        "year": [2000, 2010, 2020, 2030, 2040, 2050, 2060],
        "country": ["WORLD", "WORLD", "WORLD", "WORLD", "WORLD", "WORLD", "WORLD"],
        "imp": [
            115241.22300000001,
            109739.82599999999,
            137825.35400000002,
            148918.1680023607,
            163784.09762111588,
            174930.4527127895,
            182549.4140419694,
        ],
    }
)
irw_cz = irw.loc[{"country": "Czechia"}]
irw_world["cz_exp_1"] = (
    irw_world["imp"] * irw_cz["exp_marginal_propensity_to_export"].values
    + irw_cz["exp_constant"].values
)
irw_world["cz_exp_2"] = (
    irw_world["imp"] * irw_cz["exp_mpte_2"].values + irw_cz["exp_constant_2"].values
)


# selected_columns = df.columns
# selected_columns = selected_columns[selected_columns.str.contains("exp")]
# df_cz = df.query("country == 'Czechia'").copy()
# df_cz.plot(x="exp", y="exp_predicted", kind="scatter")

# Diagnostic plots


# Diagnostic plots
if False:
    df = irw.to_dataframe()

    # Comparison table EU Countries
    # Scatter plot to compare existing versus estimated parameters
    params = [
        "exp_marginal_propensity_to_export",
        "exp_constant",
        "exp_constant_2",
        "exp_mpte_2",
    ]
    df_params = irw[params].to_dataframe()
    df_params.plot(
        x="exp_marginal_propensity_to_export", y="exp_mpte_2", kind="scatter"
    )
    plt.show()

    # Scatter plot to compare actual versus predicted values
    df.plot(x="exp", y="exp_predicted", kind="scatter")
    plt.show()
    # Scatter plot EU Countries

    # Scatter plot to compare the world export to the sum of predicted exports
    countries = irw["country"].loc[irw.c].values
    df_agg = df.query("country in @countries").groupby("year").sum()
    df_agg.plot(x="exp", y="exp_predicted", kind="scatter")
    plt.show()

    eu_country_names = faostat.country_groups.eu_country_names
    df_eu_long = df.query("country in @eu_country_names")[
        ["exp", "exp_predicted"]
    ].melt(ignore_index=False)
    g = seaborn.relplot(
        data=df_eu_long.reset_index(),
        x="year",
        y="value",
        col="country",
        hue="variable",
        kind="line",
        col_wrap=5,
        height=3,
        facet_kws={"sharey": False, "sharex": False},
    )
    g.fig.supylabel("Actual versus predicted exports")
    g.fig.subplots_adjust(left=0.28, top=0.9)
    g.set(ylim=(0, None))
    plt.show()



"""Prepare FAOSTAT data for forest sector modelling


Note:
- The problem is that this requires the biotrade package and the biotrade_data
  directory to be defined so that the intermediate download can be saved.

"""

from functools import cached_property
from biotrade.faostat import faostat as biotrade_faostat


class FAOSTAT:
    """Download faostat data and prepare it as N dimentional Xarray data

    >>> from cobwood.faostat import faostat
    >>> fp = faostat.forestry_production_df
    >>> faostat.forestry_production_ds

    TODO: Make a dataset for sawnwood elements

    # Create a dataset
    ds = xarray.Dataset()
    selector = fp["product"] == "sawnwood"
    df = fp.loc[selector].copy()
    for element in df["element"].unique():
        selector = fp["element"] == element
        df_var = df.loc[element]
        print(element)
        # Add the data array to the dataset

    """

    def __init__(self):
        pass

    @cached_property
    def forestry_production_df(self):
        """Forestry production pandas dataframe"""
        return biotrade_faostat.pump.read_df("forestry_production")

    def foresty_production_ds(self):
        """Forestry production Xarray Dataset"""


# Make a singleton
faostat = FAOSTAT()



""" The purpose of this script is to compute the percentage increase of roundwood harvest by 2050

Run this script at the command line with:

    ipython -i ~/repos/cobwood/scripts/explore_gfpmx_data.py


- Compute the aggregate for EU countries of production and consumption.
- Compute the percentage change between any given year.

"""

import numpy as np
import pandas
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap
from biotrade.faostat import faostat
from cobwood.gfpmx_data import GFPMXData
import cobwood

gfpmx_data = GFPMXData(data_dir="gfpmx_8_6_2021")

# Path to Valerio's EU Net Annual Increment data
eu_cbm_explore_path = cobwood.data_dir.parent / "eu_cbm" / "eu_cbm_explore"
nai_path = eu_cbm_explore_path / "scenarios" / "avitabile_nai"

# Load GFTMx data
round_ = gfpmx_data.get_country_rows("round")
fuel = gfpmx_data.get_country_rows("fuel")
indround = gfpmx_data.get_country_rows("indround")
area = gfpmx_data.get_sheet_long("area")
stock = gfpmx_data.get_sheet_long("stock")

# Load NAI data and compute total yearly increment
nai_eu = pandas.read_csv(nai_path / "avitabile_2023_nai_eu.csv")
nai_eu["tot_nai_thousand_m3_ob"] = nai_eu["nai_ha"] * nai_eu["forest_area_kha"]
nai_eu["faws_nai_thousand_m3_ob"] = nai_eu["nai_ha"] * nai_eu["faws_area_kha"]
# Net Annual Increment in Forests Availaable for Wood Supply in thousand m3 under bark
nai_eu["faws_nai_thousand_m3_ub"] = nai_eu["nai_ha"] * nai_eu["faws_area_kha"] / 1.12

# Select EU countries
eucountries = faostat.country_groups.eu_country_names + ["Czech Republic"]
roundeu = round_.query("country in @eucountries")
indroundeu = indround.query("country in @eucountries")
fueleu = fuel.query("country in @eucountries")

# Check number of countries
print(set(eucountries) - set(indroundeu.reset_index().country.unique()))
print(indroundeu.reset_index().country.unique())


def print_increase_in_production(df, var, start, end):
    """Print the difference in production between the start and the end date"""
    df_change = df.loc[end, var].sum() / df.loc[start, var].sum()
    print(
        f"Increase in {df['faostat_name'].unique()} {var} between {start} and {end}:",
        round((df_change - 1) * 100),
        "%",
        f"from {round(df.loc[start, var].sum()/1e3)}",
        f"to {round(df.loc[end, var].sum()/1e3)} million m3",
    )


print_increase_in_production(roundeu, "prod", 2020, 2050)
print_increase_in_production(indroundeu, "prod", 2020, 2050)
print_increase_in_production(fueleu, "prod", 2020, 2050)

print_increase_in_production(fueleu, "prod", 2022, 2050)
print_increase_in_production(indroundeu, "prod", 2022, 2050)

print_increase_in_production(indroundeu, "cons", 2020, 2050)
print_increase_in_production(fueleu, "cons", 2020, 2050)


# Export summary tables to a spreadsheet
roundprodagg = roundeu.groupby("year")["prod"].agg(sum).rename("round")
indroundprodagg = indroundeu.groupby("year")["prod"].agg(sum).rename("indround")
fuelprodagg = fueleu.groupby("year")["prod"].agg(sum).rename("fuel")


def aggregate_eu(sheet_name, variable):
    """Load and aggregate the given variable for EU countries
    Example use:
        >>> round_prod_agg = aggregate_eu("roundprod", "prod")
        >>> stock_agg = aggregate_eu("stock", "stock")
    """
    df = gfpmx_data.get_sheet_long(sheet_name).query("country in @eucountries")
    df_agg = df.groupby("year")[variable].agg(sum).rename(sheet_name)
    return df_agg


stock_agg = aggregate_eu("stock", "stock")
area_agg = aggregate_eu("area", "area")
round_prod_agg = aggregate_eu("roundprod", "prod")
sheet_and_variable = [
    ("area", "area"),
    ("stock", "stock"),
    ("roundprod", "prod"),
    ("indroundprod", "prod"),
    ("fuelprod", "prod"),
]
agg_eu = pandas.concat(
    [aggregate_eu(*sheet_var) for sheet_var in sheet_and_variable], axis=1
)

np.testing.assert_allclose(
    agg_eu["roundprod"], agg_eu["indroundprod"] + agg_eu["fuelprod"]
)
# Note: Time series start in 1992
agg_selected = agg_eu.query("year in [1992,2000,2010,2020,2030,2050,2070]").transpose()
# Divide by 1000 and write to a csv file
# Forest area in 1000 ha -> million ha
# Stock in Million M3 -> billion m3
# Harvest in 1000 m3 -> million m3
(agg_selected / 1e3).round().to_csv("/tmp/gfpmxeuagg_selected.csv")

# Add Net Annual Increment to the harvest data
# Set the index so that it is passed on as an index
naihar_eu = agg_eu.merge(nai_eu.set_index("year"), on="year", how="left")
naihar_eu["faws_nai_thousand_m3_ub"] = naihar_eu["faws_nai_thousand_m3_ub"].interpolate(
    method="linear", limit_area="inside"
)

# Plot EU harvest projection
selector = agg_eu.index <= 2050
cols = ["roundprod", "indroundprod", "fuelprod", "nai_thousand_m3"]
cols_plot = {
    "roundprod": "Total Roundwood removals",
    "indroundprod": "Industrial Roundwood",
    "fuelprod": "Fuel wood",
    "faws_nai_thousand_m3_ub": "Net Annual Increment (FAWS, UB)",
}
(
    (naihar_eu.loc[selector, cols_plot.keys()] / 1e3)
    .rename(columns=cols_plot)
    .plot(
        title="GFPMx EU 27 harvest projections in the SSP2 scenario",
        ylabel="Million m3",
        colormap=ListedColormap(["black", "orange", "red", "green"]),
    )
)
# plt.show()
plt.savefig("/tmp/gfpmx_eu_havest.png")


# Compute production per capita
naihar_eu_agg = naihar_eu.groupby("year").sum()

# Population The EU population is projected to increase from 446.7 million in
# 2022 and peak at 453.3 million in 2026 (+1.5 %), then gradually decrease to
# 447.9 million in 2050

# Production is in 1000 m3
print(
    "EU roundwood harvest per person in 2020:",
    round(naihar_eu_agg.loc[2020, "roundprod"] / 446.7e3, 1),
    "m3",
)
print(
    "EU roundwood harvest per person in 2020:",
    round(naihar_eu_agg.loc[2050, "roundprod"] / 447.9e3, 1),
    "m3",
)
print(
    "EU faws_nai_thousand_m3_ub harvest per person in 2020:",
    round(naihar_eu_agg.loc[2020, "faws_nai_thousand_m3_ub"] / 446.7e3, 1),
    "m3",
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
The script to save all sheets of the GFPMX Excel implementation to csv files
has been moved to a function inside cobwood/gfpmx_spreadsheet_to_csv.py

Usage:

    ipython  -i  ~/repos/cobwood/scripts/gfpmx_data_to_csv.py

This script might remain useful to investigate issues with column names or sheet names.

"""

import re
import pandas

import cobwood
from cobwood.gfpmx_spreadsheet_to_csv import gfpmx_spreadsheet_to_csv

#########################################
# Convert the Excel spreadsheets to CSV #
#########################################
gfpmx_spreadsheet_to_csv("~/large_models/GFPMX-8-6-2021.xlsx")
gfpmx_spreadsheet_to_csv("~/large_models/GFPMX-base2020.xlsx")
gfpmx_spreadsheet_to_csv("~/large_models/GFPMX-base2021.xlsb")


###############################
# Investigate unnamed columns #
###############################
gfpmx_data_dir = cobwood.data_dir
gfpmx_excel_file = None

# key = "IndroundProd"
if False:
    for key in gfpmx_excel_file.keys():
        print(f"\n**{key}**")
        if key in ["FuelProd$", "IndroundProd$"]:
            print("lala")
        df = gfpmx_excel_file[key]
        # Those operations are duplicated from above
        # Remove empty columns in place
        df.dropna(how="all", axis=1, inplace=True)
        # Rename columns to snake case, replace all non alphanumeric characters by an underscore
        df.rename(columns=lambda x: re.sub(r"\W+", "_", str(x)).lower(), inplace=True)
        # Rename unnamed columns if they have unique values
        unnamed = df.filter(regex="unnamed").columns.to_list()
        for col in unnamed:
            content = df[col].dropna()
            new_name = col
            if len(content.unique()) != 1 or not pandas.api.types.is_string_dtype(
                content
            ):
                break
            if content.str.contains("m3").all():
                new_name = "unit"
            if content.str.contains(r"[^\W\d_]").all():
                new_name = "element"
            print(
                f"Old name: '{col}' content: {content.unique()}, new name: '{new_name}'"
            )



"""Plot GFPMx dataset objects by region (continent) or by country

"""

import seaborn
import xarray


def plot_da_by_region(
    ds: xarray.Dataset, da_name: str
) -> xarray.plot.facetgrid.FacetGrid:
    """Plot a data array by region given a dataset and the name of the data array inside it.

    Example:

        >>> from cobwood.gfpmx import GFPMX
        >>> from cobwood.gfpmx_plot import plot_da_by_region
        >>> gfpmxb2021 = GFPMX(
        ...     input_dir="gfpmx_base2021", base_year=2021, scenario="base_2021", rerun=False
        ... )
        >>> plot_da_by_region(gfpmxb2021["indround"], "prod")

    """
    return ds[da_name].loc[~ds.c].plot(col="country", col_wrap=4)


def facet_plot_by_var(
    ds: xarray.Dataset,
    variables: list = None,
    countries: list = None,
    ylabel: str = None,
    title: str = None,
) -> seaborn.axisgrid.FacetGrid:
    """Plot the given dataset variables with a facet for each variable and a
    color line for each continent

    Example:

        >>> from cobwood.gfpmx import GFPMX
        >>> from cobwood.gfpmx_plot import facet_plot_by_var
        >>> gfpmxb2021 = GFPMX(
        >>>     input_dir="gfpmx_base2021", base_year=2021, scenario="base_2021", rerun=False
        >>> )
        >>> # By default plot one line by continent
        >>> facet_plot_by_var(gfpmxb2021.indround)
        >>> # The country argument can specify one line by country
        >>> facet_plot_by_var(gfpmxb2021.indround, countries=["Canada", "France", "Japan"])
        >>> # Plot Forest area and forest stock
        >>> facet_plot_by_var(gfpmxb2021.other, ["area", "stock"],
        >>>                  ylabel="Area in 1000ha and stock in million m3")

    """
    if variables is None:
        variables = ["cons", "imp", "exp", "prod", "price"]
    if ylabel is None:
        ylabel = "Quantity in 1000 m3, price in USD/m3"
    if title is None:
        title = ds.product
    # Keep only the selected variables
    if countries is None:
        # Select continents
        df = ds.loc[{"country": ~ds.c}][variables].to_dataframe()
    else:
        # Select countries
        df = ds.loc[{"country": countries}][variables].to_dataframe()
    df = df.reset_index().melt(id_vars=["country", "year"])
    g = seaborn.relplot(
        data=df,
        x="year",
        y="value",
        col="variable",
        hue="country",
        kind="line",
        col_wrap=5,
        height=3,
        facet_kws={"sharey": False, "sharex": False},
    )
    g.set(ylim=(0, None))
    g.fig.supylabel(ylabel)
    g.fig.suptitle(title)
    g.fig.subplots_adjust(left=0.09, top=0.85)
    return g



"""This scripts computes all GFPMx equations using xarray

TODO: Continue work in scripts/compute_gdp_fair_scenario.py

Usage:

    ipython -i ~/repos/cobwood/scripts/compute_all_equations_with_xarray.py

See also:

- The general version of this script in the module cobwood/run_and_compare_to_ref.py

- A drawing illustrating the model structure in
  draft_articles/gftm_degrowth/model_structure.odg

- Buongiorno, J. (2021). GFPMX: A Cobweb Model of the Global Forest Sector,
  with an Application to the Impact of the COVID-19 Pandemic. Sustainability,
  13(10), 5507. https://www.mdpi.com/2071-1050/13/10/5507

Advantages of using xarray over pandas:

- The shift_index() function is not needed any more, we can call
  sawn["price"].loc[:, t - 1] directly.


- Evaluate whether it is better to keep countries and aggregates in the same dataset.

- TODO within option A replace COUNTRIES by a list internal to the dataset for example

  >>> ds = sawn
  >>> all(ds["price"].loc[~ds.region.isnull(), 2019] ==  ds["price"].loc[COUNTRIES, 2019])
  >>> ds["c"] = ~ds.region.isnull()
  >>> ds["price"].loc[ds.c, 2019]
  >>> assert(all(ds["country"].loc[~ds.c] ==
  >>>            ['WORLD', 'AFRICA', 'NORTH AMERICA', 'SOUTH AMERICA', 'ASIA', 'OCEANIA', 'EUROPE']))


  Option A first option as of begining of 2023. Countries and aggregates are in the same dataset, consumption can be
  computed for countries only first, the aggregates computed later.

        ds["cons_constant"]
        * pow(ds["price"].loc[COUNTRIES, t - 1], ds["cons_price_elasticity"])
        * pow(ds["gdp"].loc[COUNTRIES, t], ds["cons_gdp_elasticity"])

  World price is

        ds["price_constant"].loc["WORLD"]
        * pow(ds_primary["price"].loc["WORLD", t], ds["price_input_elast"].loc["WORLD"])

  Should the .loc["WORLD"] and .loc[COUNTRIES] be done when the function is
  called? No better to leave no ambiguity.

  All uses of the world aggregates in the equations below, show that only the
  world price, world production and world stock variables are used in equations:

        ds["price_constant"].loc["WORLD"]
        pow(ds_primary["price"].loc["WORLD", t], ds["price_input_elast"].loc["WORLD"])
        ds_agg["price_constant"].loc["WORLD"]
        pow(ds_primary_agg["price"].loc["WORLD", t], ds_agg["price_input_elast"].loc["WORLD"])
        ds["price_constant"].loc["WORLD"]
        ds_primary["price"].loc["WORLD", t], ds["price_input_elast"].loc["WORLD"]
        ds["price_constant"].loc["WORLD"]
        ds["prod"].loc["WORLD", t],
        ds["price_world_price_elasticity"].loc["WORLD"],
        ds_other["stock"].loc["WORLD", t],
        ds["price_stock_elast"].loc["WORLD"],
        np.exp(ds["price_trend"].loc["WORLD"] * t)
        ds["price"].loc["WORLD", t], ds["price_world_price_elasticity"].loc[ds.c]


  Option B There is a dataset of countries only and a dataset of aggregates only we
  can compute consumption as such:

        ds["cons_constant"]
        * pow(ds["price"].loc[:, t - 1], ds["cons_price_elasticity"])
        * pow(ds["gdp"].loc[:, t], ds["cons_gdp_elasticity"])

  World prices can be computed as:

        ds_agg["price_constant"].loc["WORLD"]
        * pow(ds_primary_agg["price"].loc["WORLD", t], ds_agg["price_input_elast"].loc["WORLD"])

  Option C Aggregates are stored as another dimension. Since the continent
  aggregates are not used, they could be stored separately as part of the
  dataset. Or the dataset could contain another dimension that distinguishes
  the aggregates.


    Do we need a big or a small class? Do we need a class at all?
    - Con: This would mean using methods instead of functions. Nah.
       - All the computation functions can remain as functions, they don't need to be part of the class
    - Pro: a class would mean storing all data sets as class attributes,
      and providing an object for model comparison yes.
          - The only blocker is the COUNTRIES list, which is also part of gfpmx_data.

    class GFPMX_runner:
        # Run the GFPMX simulation, I"m not in favour of this big monster class
        def __init__(self, data_dir, base_year):
            self.gfpmx_data = GFPMXData(data_dir=data_dir, base_year=base_year)

    class mini_gfpmx_data:
       # carry only indround, fuel, pulp, sawn, panel, paper

  Option D Aggregates are only available as a method computed from the data, based on the


"""


from cobwood.gfpmx_data import convert_to_2d_array
from cobwood.gfpmx_data import GFPMXData
from cobwood.gfpmx_data import remove_after_base_year_and_copy
from cobwood.gfpmx_data import compare_to_ref
from cobwood.gfpmx_equations import compute_one_time_step
from cobwood.gfpmx_equations import consumption
from cobwood.gfpmx_equations import import_demand
from cobwood.gfpmx_equations import export_supply
from cobwood.gfpmx_equations import production
from cobwood.gfpmx_equations import world_price
from cobwood.gfpmx_equations import local_price

BASE_YEAR = 2018
DATA_DIR = "gfpmx_8_6_2021"
gfpmx_data = GFPMXData(data_dir=DATA_DIR, base_year=BASE_YEAR)

# Load reference data
other_ref = gfpmx_data.convert_sheets_to_dataset("other")
indround_ref = gfpmx_data.convert_sheets_to_dataset("indround")
round_ref = gfpmx_data.convert_sheets_to_dataset("round")
fuel_ref = gfpmx_data.convert_sheets_to_dataset("fuel")
sawn_ref = gfpmx_data.convert_sheets_to_dataset("sawn")
panel_ref = gfpmx_data.convert_sheets_to_dataset("panel")
pulp_ref = gfpmx_data.convert_sheets_to_dataset("pulp")
paper_ref = gfpmx_data.convert_sheets_to_dataset("paper")
gdp = convert_to_2d_array(gfpmx_data.get_sheet_wide("gdp"))

COUNTRY_AGGREGATES = gfpmx_data.country_aggregates
COUNTRIES = sawn_ref.country[~sawn_ref.country.isin(COUNTRY_AGGREGATES)].values
# Check that all countries in the dataset are also present in the list
assert set(list(COUNTRIES)) - set(gfpmx_data.country_groups["country"]) == set()
# Check that all countries present in the list are also in the dataset
assert set(gfpmx_data.country_groups["country"]) - set(list(COUNTRIES)) == set()

######################################################
# Erase data after base year and copy the data frame #
######################################################
# Using a mask makes the dataset return an error on loc selection
# base_year = 2018
# mask = sawn.coords["year"] > base_year
# sawn = xarray.where(mask, np.nan, sawn_ref).copy()
# sawn["price"].loc[:,t-1]
# # returns KeyError: 2018
# # But selecting the original data works fine
# sawn_ref["price"].loc[:,t-1]
# --> Make a reproducible example and ask on Stackoverflow why this is the case.
# We will compute demand from the base_year + 1

other = remove_after_base_year_and_copy(other_ref, BASE_YEAR)
fuel = remove_after_base_year_and_copy(fuel_ref, BASE_YEAR)
indround = remove_after_base_year_and_copy(indround_ref, BASE_YEAR)
# Use an underscore so that we don't overwrite the python round() function
round_ = remove_after_base_year_and_copy(round_ref, BASE_YEAR)
sawn = remove_after_base_year_and_copy(sawn_ref, BASE_YEAR)
panel = remove_after_base_year_and_copy(panel_ref, BASE_YEAR)
pulp = remove_after_base_year_and_copy(pulp_ref, BASE_YEAR)
paper = remove_after_base_year_and_copy(paper_ref, BASE_YEAR)

# Memory size
for ds in [indround, fuel, pulp, sawn, panel, paper, other]:
    print(f"{ds.product}: {ds.nbytes / 1024 ** 2:.2f} MB")

# Add GDP projections to the datasets gdp are projected to the future
sawn["gdp"] = gdp
round_["gdp"] = gdp
panel["gdp"] = gdp
fuel["gdp"] = gdp
paper["gdp"] = gdp

compute_one_time_step(indround, fuel, pulp, sawn, panel, paper, other, 2019)
compute_one_time_step(indround, fuel, pulp, sawn, panel, paper, other, 2020)
# See more years below, using the compare_to_ref function to compare

# # Compute roundwood as the sum of industrial round wood and fuel wood
# round_["prod"].loc[COUNTRIES, year] = (
#     indround["prod"].loc[COUNTRIES, year] + fuel["prod"].loc[COUNTRIES, year]
# )

##########
# Checks #
##########


compute_one_time_step(indround, fuel, pulp, sawn, panel, paper, other, 2019)
ciepp_vars = ["cons", "imp", "exp", "prod", "price"]
compare_to_ref(sawn, sawn_ref, ciepp_vars, 2019)
compare_to_ref(panel, panel_ref, ciepp_vars, 2019)
compare_to_ref(paper, paper_ref, ciepp_vars, 2019)
compare_to_ref(pulp, pulp_ref, ciepp_vars, 2019)
compare_to_ref(indround, indround_ref, ciepp_vars, 2019)

compute_one_time_step(indround, fuel, pulp, sawn, panel, paper, other, 2020)
ciepp_vars = ["cons", "imp", "exp", "prod", "price"]
compare_to_ref(sawn, sawn_ref, ciepp_vars, 2020)
compare_to_ref(panel, panel_ref, ciepp_vars, 2020)
compare_to_ref(paper, paper_ref, ciepp_vars, 2020)
compare_to_ref(pulp, pulp_ref, ciepp_vars, 2020)
compare_to_ref(indround, indround_ref, ciepp_vars, 2020)

for this_year in range(BASE_YEAR + 1, 2051):
    print(this_year)
    # TODO: decrease tolerance
    rtol = 1e-2
    compute_one_time_step(indround, fuel, pulp, sawn, panel, paper, other, this_year)
    ciepp_vars = ["cons", "imp", "exp", "prod", "price"]
    compare_to_ref(sawn, sawn_ref, ciepp_vars, this_year, rtol=rtol)
    compare_to_ref(panel, panel_ref, ciepp_vars, this_year, rtol=rtol)
    compare_to_ref(paper, paper_ref, ciepp_vars, this_year, rtol=rtol)
    compare_to_ref(pulp, pulp_ref, ciepp_vars, this_year, rtol=rtol)
    compare_to_ref(indround, indround_ref, ciepp_vars, this_year, rtol=rtol)

# Compare world price
#  assert_allclose(sawn["price"].loc["WORLD", year], sawn_ref["price"].loc["WORLD", year])

raise ValueError(
    """ds["price"].loc[ds.c, 2019] opens the possibility to use the class gfpmx_data
                 as the sole argument of compute_one_time_step"""
)


year = 2019
# Comparison between sawn modified in place by the
# compute_secondary_product_ciep() function
# and sawn2 computed below
sawn2 = sawn.copy(deep=True)
sawn2["cons"].loc[COUNTRIES, year] = consumption(sawn2, year)
sawn2["imp"].loc[COUNTRIES, year] = import_demand(sawn2, year)
sawn2["exp"].loc[COUNTRIES, year] = export_supply(sawn2, year)
sawn2["prod"].loc[COUNTRIES, year] = production(sawn2, year)
sawn2["price"].loc["WORLD", year] = world_price(sawn2, indround, year)
sawn2["price"].loc[COUNTRIES, year] = local_price(sawn2, year)
assert sawn.equals(sawn2)


# Comparison in case of mismatch
sawn_df = sawn.loc[dict(country=COUNTRIES, year=year)].to_pandas().reset_index()
sawn_ref_df = sawn_ref.loc[dict(country=COUNTRIES, year=year)].to_pandas().reset_index()
cols_to_check = ["cons", "imp", "exp", "prod"]
print(sawn_df[cols_to_check])
print(sawn_ref_df[cols_to_check])
sawn_df["prod_diff"] = sawn_df["prod"] / sawn_ref_df["prod"] - 1
sawn_df["cons_diff"] = sawn_df["cons"] / sawn_ref_df["cons"] - 1
print(sawn_df[["cons_diff", "prod_diff"]].describe())
sawn_df[["country"] + cols_to_check + ["prod_diff"]].sort_values(
    "prod_diff", ascending=False
)
print(sawn_ref_df[["country"] + cols_to_check])


# Plot comparison
# sawn_df["reference"] = "no"
# sawn_ref_df["reference"] = "yes"
# sawn_comp = pandas.concat([sawn_df, sawn_ref_df])
# p = seaborn.scatterplot(x="country", y="prod", hue="reference", data=sawn_comp)
# plt.xticks(rotation=90)
# plt.show()



"""
Load PIK-Magpie scenario data from the paper

- Bodirsky, B.L., Chen, D.MC., Weindl, I. et al. Integrating degrowth and efficiency
  perspectives enables an emission-neutral food system by 2100. Nat Food 3, 341–348
  (2022). https://doi.org/10.1038/s43016-022-00500-3

Run this script at the command line with:

    ipython -i ~/repos/cobwood/scripts/load_pik_data.py

See also the associated notebook with comparison plots:

    ../notebooks/explore_pik_gdp_scenarios.md

The data is located at:

- https://zenodo.org/record/5543427#.Y3eYOkjMKcM

The readme says "unit: population: Mio people, gdp: Million USD,"

The data in the figures/scenariogdx folder is in the form of GAMMS GDX files
(although the model interface seems to be in R). There are also csv files in
the figures/incomes folder.


# Mail to Bodirsky

Dear Benjamin Bodirsky,

I enjoyed reading your paper "Integrating degrowth and efficiency perspectives
enables an emission-neutral food system by 2100". I am implementing a forest
sector model which has wood consumption driven by a revenue elasticity of
demand (where revenue is proxied by the GDP). The GFPMx model was published as
"GFPMX: A Cobweb Model of the Global Forest Sector, with an Application to the
Impact of the COVID-19 Pandemic"


I would like to use the GDP projections from your BAU and FAIR GDP scenarios.
Based on the data at https://zenodo.org/record/5543427#.Y3eYOkjMKcM , I
compared the value in DegrowthMAgPIE/figures/incomes/bau_gdp_ppp_iso.csv to
historical GDP values from the World Bank
https://data.worldbank.org/indicator/NY.GDP.MKTP.PP.KD and to values from the
forest sector model GFPMx (also based on the SSP2 storyline). Both other
sources are expressed in constant USD of 2017. See comparison plot attached for
EU countries only. I am wondering why the BAU SSP2 GDP values are almost always
below the world bank values for the historical period. I am also wondering
about the future projection period with the GFPMx model, but I guess I would
have to check

I was trying to look at the FAIR scenario, but it seems that the values in
fair_gdp_ppp_iso.csv do not change much. Do I need to look at the SSP2 column
in that file or at another column?  Where can I look for the GDP projections
used in the FAIR scenario?

Kind regards,
Paul Rougieux


# Reply from co author David M Chen

> "Yes, our GDP units are in constant 2005 USD at market-exchange rate, which
> should explain the consistent difference between them and the World Bank 2017
> values. Our historic GDP values are based on a harmonized dataset (that
> primarily also uses WB as source) documented here, for which we have an updated
> version from 2019.

> And yes for the FAIR scenario, that's a bit unclear, sorry about that, the
> SSP2 column in fair_gdp_ppp_iso.csv has the per capita GDP projections for the
> FAIR scenario. The other columns are the other SSP scenarios as they are (they
> have not been subject to the FAIR redistribution)."

"""
# TODO: update to 2017 constant usd
# Search "world bank change constant USD reference year"

from pathlib import Path
import re

import pandas

# To get country ISO codes
from biotrade.faostat import faostat
from biotrade.world_bank import world_bank

# import gdxpds
import cobwood
from cobwood.gfpmx_data import GFPMXData


country_iso_codes = faostat.country_groups.df[["fao_table_name", "iso3_code"]].rename(
    columns={"fao_table_name": "country", "iso3_code": "country_iso"}
)
# Correct mapping of China and Netherlands to the names used in GFPMX
country_iso_codes = (
    country_iso_codes.query("country != 'China'")  # Remove existing China
    .copy()
    .replace("China, mainland", "China")
    .replace("Netherlands (Kingdom of the)", "Netherlands")
)

# Requires a working version of GAMMS
# dataframes = gdxpds.to_dataframes(magpie_scenario_path / "bau_fulldata.gdx")
# for symbol_name, df in dataframes.items():
#     print("Doing work with {}.".format(symbol_name))

############################
# Load World Bank GDP data #
############################
# Load World Bank GDP, PPP (constant 2017 international $)
# https://data.worldbank.org/indicator/NY.GDP.MKTP.PP.KD
wb_gdp_cst = world_bank.db.select("indicator", indicator_code="NY.GDP.MKTP.PP.KD")
wb_gdp_cst.rename(
    columns={"reporter_code": "country_iso", "value": "wb_gdp_cst"}, inplace=True
)
# Convert from USD to million USD
wb_gdp_cst["wb_gdp_cst"] = wb_gdp_cst["wb_gdp_cst"] / 1e6

# Load World Bank GDP, PPP (current international $)
wb_gdp_cur = world_bank.db.select("indicator", indicator_code="NY.GDP.MKTP.PP.CD")
wb_gdp_cur.rename(
    columns={"reporter_code": "country_iso", "value": "wb_gdp_cur"}, inplace=True
)
# Convert from USD to million USD
wb_gdp_cur["wb_gdp_cur"] = wb_gdp_cur["wb_gdp_cur"] / 1e6

index = ["country_iso", "reporter", "year"]
wb = wb_gdp_cst[index + ["wb_gdp_cst"]].merge(wb_gdp_cur[index + ["wb_gdp_cur"]])

# Check that the constant 2017 USD values correspond to the current value in 2017
# wb["i"] = wb["wb_gdp_cst"]
wb.query("year == 2017 and reporter in @faostat.country_groups.eu_country_names")
# Keep non empty values
wb2017 = wb.query("year == 2017 and wb_gdp_cst == wb_gdp_cst").copy()
# np.testing.assert_allclose(wb2017["wb_gdp_cst"], wb2017["wb_gdp_cur"])
wb2017["diff"] = wb2017["wb_gdp_cst"] - wb2017["wb_gdp_cur"]
wb2017.query("abs(diff) > 1")
# Values are mostly equals except in some special groupings

###############################
# Load PIK Magpie income data #
###############################
scenario_path = Path("~/large_models/DegrowthMAgPIE/figures/scenariogdx")
incomes_path = Path("~/large_models/DegrowthMAgPIE/figures/incomes")


def load_pik_gdp(file_path, new_var_name):
    """Load a pik GDP file and rename the SSP2 column to a new name"""
    df = pandas.read_csv(file_path)
    df.rename(columns=lambda x: re.sub(r" ", "_", str(x)).lower(), inplace=True)
    df.rename(columns={"region": "country_iso", "ssp2": new_var_name}, inplace=True)
    df["year"] = pandas.to_numeric(df["year"].str.replace("y", ""))
    return df


bau_gdp = load_pik_gdp(incomes_path / "bau_gdp_ppp_iso.csv", "pik_bau")
fair_gdp = load_pik_gdp(incomes_path / "fair_gdp_ppp_iso.csv", "pik_fair")


#######################
# Load GFPMX GDP data #
#######################
# Expressed in 1000 USD of 2017

gfpmx_data_b2018 = GFPMXData(data_dir="gfpmx_8_6_2021")
gfpmx_data_b2021 = GFPMXData(data_dir="gfpmx_base2021")


def load_gfpmx_gdp(gfpmx_data: GFPMXData, col_name):
    """Load GFPMX GDP data and give it a column name"""
    df = gfpmx_data.get_sheet_long("gdp").merge(
        country_iso_codes, on="country", how="left"
    )
    # Convert from 1000 USD to million USD
    df["gdp"] = df["gdp"] / 1e3
    df.rename(columns={"gdp": col_name}, inplace=True)
    df.drop(columns="gdp_unnamed_0", inplace=True)
    return df


index = ["country", "country_iso", "year"]
gfpm_gdp = load_gfpmx_gdp(gfpmx_data_b2018, "gfpm_gdp_b2018").merge(
    load_gfpmx_gdp(gfpmx_data_b2021, "gfpm_gdp_b2021"), on=index, how="outer"
)

# Divide
# wb["i"] = wb["wb_gdp_cst"] /
# df['normal'] = df.Value / df['VALUE'].where(df.TIME.str[5:] =='Q1').groupby(df['LOCATION']).transform('first')


# Max GDP per region
bau_gdp.loc[bau_gdp.groupby("country_iso")["pik_bau"].idxmax()]
bau_gdp.query("country_iso=='FRA'")

# Check that there are no NA values for EU country ISO codes
selector = gfpm_gdp["country"].isin(faostat.country_groups.eu_country_names)
assert not any(gfpm_gdp[selector]["country_iso"].isna())


##############################
# Merge data frames together #
##############################
# Merge GFPM, Magpie and world bank data and compare the GDP of EU countries in 2030
index = ["country_iso", "year"]
comp = (
    gfpm_gdp[index + ["country", "gfpm_gdp_b2018", "gfpm_gdp_b2021"]]
    .merge(bau_gdp[index + ["pik_bau"]], on=index, how="left")
    .merge(fair_gdp[index + ["pik_fair"]], on=index, how="left")
    .merge(wb[index + ["wb_gdp_cst", "wb_gdp_cur"]], on=index, how="left")
)

# Reshape to long format
comp_long = comp.melt(
    id_vars=["country_iso", "year", "country"], var_name="source", value_name="gdp"
)

###############################################
# Rescale PIK to a 2017 or a 2021 base year # #
###############################################
#
# https://datahelpdesk.worldbank.org/knowledgebase/articles/114946-how-can-i-rescale-a-series-to-a-different-base-yea
#
# > "For example, you can rescale the 2010 data to 2005 by first creating an index
# > dividing each year of the constant 2010 series by its 2005 value (thus, 2005 will
# > equal 1). Then multiply each year's index result by the corresponding 2005 current
# > U.S. dollar price value."
#
# PIK GDP values are in constant 2005 USD
# Create an index based on the 2017 value

# Convert from 2017 to 2017 values
# - There is no data in 2017, need to interpolate

comp["pik_bau_i"] = comp.groupby("country_iso")["pik_bau"].transform(
    pandas.Series.interpolate, "linear"
)
comp["pik_fair_i"] = comp.groupby("country_iso")["pik_fair"].transform(
    pandas.Series.interpolate, "linear"
)
column_names_2017 = {
    "gfpm_gdp_b2018": "gfpm_gdp_2017",
    "wb_gdp_cur": "wb_gdp_2017",
    "pik_bau_i": "pik_bau_2017",
    "pik_fair_i": "pik_fair_2017",
}
gdp_2017 = (
    comp.loc[comp["year"] == 2017, ["country_iso", *column_names_2017.keys()]]
    .rename(columns=column_names_2017)
    # Remove NA values in country
    .query("country_iso == country_iso")
    .copy()
)
column_names_2021 = {
    "gfpm_gdp_b2021": "gfpm_gdp_2021",
    "pik_bau_i": "pik_bau_2021",
    "pik_fair_i": "pik_fair_2021",
}
gdp_2021 = (
    comp.loc[comp["year"] == 2021, ["country_iso", *column_names_2021.keys()]]
    .rename(columns=column_names_2021)
    .query("country_iso == country_iso")
    .copy()
)

code_continent = faostat.country_groups.df[["iso3_code", "continent"]].rename(
    columns={"iso3_code": "country_iso"}
)
index = ["country_iso", "year"]
gdp_comp = (
    comp.merge(gdp_2017, on="country_iso")
    .merge(gdp_2021, on="country_iso")
    .assign(
        # Scale PIK GDP to World bank GDP 2017
        pik_bau_adjwb2017=lambda x: (x["pik_bau_i"] / x["pik_bau_2017"])
        * x["wb_gdp_2017"],
        pik_fair_adjwb2017=lambda x: (x["pik_fair_i"] / x["pik_fair_2017"])
        * x["wb_gdp_2017"],
        # Scale PIK GDP to GFPMX GDP 2017
        pik_bau_adjgfpm2017=lambda x: (x["pik_bau_i"] / x["pik_bau_2017"])
        * x["gfpm_gdp_2017"],
        pik_fair_adjgfpm2017=lambda x: (x["pik_fair_i"] / x["pik_fair_2017"])
        * x["gfpm_gdp_2017"],
        # Scale PIK GDP to GFPMX GDP 2021
        pik_bau_adjgfpm2021=lambda x: (x["pik_bau_i"] / x["pik_bau_2021"])
        * x["gfpm_gdp_2021"],
        pik_fair_adjgfpm2021=lambda x: (x["pik_fair_i"] / x["pik_fair_2021"])
        * x["gfpm_gdp_2021"],
    )
    .drop(columns=list(column_names_2017.values()) + list(column_names_2021.values()))
    .merge(code_continent, on="country_iso")
)

# Shift by 5 years
gdp_comp["pik_fair_shift_5"] = gdp_comp.groupby("country_iso")[
    "pik_fair_adjgfpm2021"
].shift(periods=5)

# Reshape to long format
index = ["country_iso", "year", "country", "continent"]
gdp_comp_long = gdp_comp.drop(columns=["pik_bau_i", "pik_fair_i"]).melt(
    id_vars=index, var_name="source", value_name="gdp"
)

# See plot comparison between rescaled and original data in the notebook


# Create data frames for EU countries only
selector = comp["country"].isin(faostat.country_groups.eu_country_names)
comp_eu = comp[selector].copy()
selector = comp_long["country"].isin(faostat.country_groups.eu_country_names)
comp_eu_long = comp_long[selector].copy()

# Check that the PIK constant 2005 USD values correspond to WB current value in 2005
comp_eu_2005 = comp_eu.query("year == 2005 and pik_bau == pik_bau")


############################################
# Rescale using the rate of growth instead #
############################################
# What countries are available and how does is the fair redistribution implemented?
# For each country compute the proportion decrease between 2020 and 2030.


###################################################
# Keep only selected year for comparison purposes #
###################################################
eu_countries = faostat.country_groups.eu_country_names
gdp_comp_eu = gdp_comp.query("country in @faostat.country_groups.eu_country_names")
# Keep only ssp2 and adjusted pik fair scenario
# Keep country iso to check we only have EU countries
agg = {"country_iso": lambda x: "".join(x.unique()), "gdp": "sum"}
gdp_comp_long_agg_eu_2 = (
    gdp_comp.melt(
        id_vars=["country_iso", "year", "country", "continent"],
        var_name="source",
        value_name="gdp",
    )
    .query("country in @eu_countries")
    .groupby(["year", "source"])
    .agg(agg)
    .reset_index()
    # TODO: fix this in a more elegant way
    .query("gdp>0.1")
    .copy()
)
print(gdp_comp_long_agg_eu_2["country_iso"].unique())
gdp_comp_long_agg_eu_2.value_counts(["source", "country_iso"])

####################
# Aggregate EU GDP #
####################
# For a summary table in the technical report
selected_sources = ["gfpm_gdp_b2021", "pik_fair_adjgfpm2021", "pik_bau_adjgfpm2021"]
selected_years = [1992, 2000, 2010, 2020, 2030, 2050, 2070]
gdp_comp_eu_selected = (
    gdp_comp_long_agg_eu_2
    # Convert from million USD to billion USD
    .assign(gdp_bil=lambda x: x["gdp"] / 1e3)
    .query("year in @selected_years and source in @selected_sources")
    .pivot(index="source", columns="year", values="gdp_bil")
)
# Year 1992 is only available in GFPM base 2018
# Compare GFPM GDP base 2018 and base 2021
gdp_b2018_b2021 = (
    gdp_comp_long_agg_eu_2.query("source in ['gfpm_gdp_b2018', 'gfpm_gdp_b2021']")
    # Keep this index on year please it is needed for the plot and the loc selection
    .pivot(columns="source", index="year", values="gdp")
)
# Linear interpolation of 1992
gdp_b2018_b2021["gfpm_gdp_b2021_i"] = gdp_b2018_b2021["gfpm_gdp_b2021"].interpolate(
    method="linear", limit_direction="both"
)
# gdp_b2018_b2021.plot()
gdp_comp_eu_selected.loc["gfpm_gdp_b2021", "1992"] = (
    float(gdp_b2018_b2021.loc[1992, "gfpm_gdp_b2021_i"]) / 1e3
)


gdp_comp_eu_selected.round().to_csv("/tmp/gdp_comp_eu_selected.csv")

# Use a simple ratio to compute back GFPM GDP 1992 from GFPMx GDP b 2018
gdp_comp_long_agg_eu_2.query("year in [1992, 2000]")


#####################
# Store output data #
#####################
pik_data_dir = cobwood.create_data_dir("pik")

# Write EU GDP to a parquet file
(
    # Remove interpolated data
    comp_eu.drop(columns=["pik_bau_i", "pik_fair_i"]).to_parquet(
        pik_data_dir / "comp_eu.parquet"
    )
)

# Write GDP comparison to a parquet file
gdp_comp.to_parquet(pik_data_dir / "gdp_comp.parquet")

# Write to a compressed csv file
compression_opts = dict(method="zip", archive_name="comp.csv")
(
    comp.drop(columns=["pik_bau_i", "pik_fair_i"]).to_csv(  # Remove interpolated data
        "/tmp/comp.csv.zip", index=False, compression=compression_opts
    )
)



"""Load Eurostat population data

Load historical population data, load population projections, merge them together.

Sources:

    - Historical data
    https://ec.europa.eu/eurostat/web/population-demography/demography-population-stock-balance/database

    - Population projections
    https://ec.europa.eu/eurostat/web/population-demography/population-projections/database

Usage:

    >>> from cobwood.eurostat.population import load_hist_population_and_proj
    >>> popt = load_hist_population_and_proj("eurostat_tps00001_page.tsv", "eurostat_proj_23np.tsv")

Separate low level functions for information and debugging:

    >>> from cobwood.eurostat.population import load_eurostat
    >>> from cobwood.eurostat.population import combine_hist_population_to_proj
    >>> # Load Eurostat historical population
    >>> from cobwood.eurostat.population import load_hist_population_and_proj
    >>> pop_hist = load_eurostat("eurostat_tps00001_page.tsv")
    >>> # Load Eurostat population projection
    >>> pop_proj_23 = load_eurostat("eurostat_proj_23np.tsv")
    >>> # combine them in one data frame
    >>> popt = combine_hist_population_to_proj(pop_hist, pop_proj_23)
    >>> # Descriptive statistics of the projections
    >>> for col in ["freq", "projection", "sex", "age", "unit", "geo", "year"]:
    >>>     print(col, ":", pop_proj_23[col].unique())

Plot the historical population data and baseline projection

    >>> from matplotlib import pyplot as plt
    >>> popt.set_index("year").plot()
    >>> plt.show()

Plot all projections

    >>> selector = pop_proj_23["geo"] == "EU27_2020"
    >>> selector &= pop_proj_23["sex"] == "T"
    >>> pop_proj_23["value_m"] =  pop_proj_23["value"] / 1e6
    >>> proj_wide = pop_proj_23.loc[selector].pivot(columns="projection", index="year", values="value_m")
    >>> proj_wide.plot(title="EU population projection in million inhabitants")
    >>> plt.show()

"""

import pandas
from pandas.api.types import is_string_dtype
from cobwood import cobwood_data_dir


def reformat_eurostat(df):
    """Reformat the first columns of a eurostat tab separated value dataset
    which are actually comma separated values, reshape year in long format."""
    raw_first_col = df.columns[0]
    col_names = df.columns.str.replace("\\TIME_PERIOD", "")[0].split(",")
    df[col_names] = df[raw_first_col].str.split(",", expand=True)
    df.drop(columns=raw_first_col, inplace=True)
    # Clean column names
    df.columns = df.columns.str.replace(" ", "")
    # Reshape to long format
    df_long = df.melt(id_vars=col_names, var_name="year", value_name="value")
    df_long["year"] = df_long["year"].astype(int)
    # If value is a string column, separate the flag to a new column
    if is_string_dtype(df_long["value"]):
        df_long[["value", "flag"]] = df_long["value"].str.split(" ", n=1, expand=True)
    return df_long


def load_eurostat(file_name):
    """Load and reformat a Eurostat data frame"""
    df = pandas.read_table(cobwood_data_dir / "eurostat" / file_name)
    df = reformat_eurostat(df)
    return df


def combine_hist_population_to_proj(df_hist, df_proj):
    """Prepare population projection by interpolating missing years in the
    projection (which has a 5 years interval) and adding data for the
    historical period. df_hist contains the historical data, df_proj contains
    the projection data."""
    # Combine historical data and projection
    year_max_hist = int(df_hist["year"].max())
    # Keep only total EU population
    selector_proj = df_proj["geo"] == "EU27_2020"
    # BSL is the baseline projection
    selector_proj &= df_proj["projection"] == "BSL"
    selector_proj &= df_proj["sex"] == "T"
    selector_proj &= df_proj["year"] > year_max_hist
    selector_hist = df_hist["geo"] == "EU27_2020"
    cols = ["freq", "geo", "year", "value"]
    df = pandas.concat(
        [df_hist.loc[selector_hist, cols], df_proj.loc[selector_proj, cols]],
        ignore_index=True,
    )
    # The projection is on a 5 years interval.
    # Interpolate between missing years.
    more_years = pandas.DataFrame({"year": range(min(df["year"]), max(df["year"]))})
    df = df.merge(more_years, on="year", how="outer")
    df.sort_values("year", inplace=True)
    df["value"] = pandas.to_numeric(df["value"], errors="coerce")
    df["value"] = df["value"].interpolate(method="linear")
    df["freq"] = df["freq"].ffill()
    df["geo"] = df["geo"].ffill()
    return df


def load_hist_population_and_proj(file_hist, file_proj):
    """Load historical population and population projection total EU"""
    pop_hist = load_eurostat(file_hist)
    pop_proj = load_eurostat(file_proj)
    # combine them in one data frame
    df = combine_hist_population_to_proj(pop_hist, pop_proj)
    return df



"""Reimplement the GFPMx model with Xarray

Usage:

    >>> from cobwood.gfpmx_equations import compute_one_time_step

"""

import numpy as np
import xarray
import pandas

# Reproduce behaviour/bugs in GFPMX-8-6-2021.xlsx
GFPMX_8_6_2021_COMPATIBLE_MODE = True
# List of bugs:
# 1) Indonesia has negative industrial roundwood consumption because
#    =IF($PulpProd.AK118+$PanelProd.AK118+$SawnProd.AK118<=0,0,$G118*($IndroundPrice.AJ118^$D118)*($PulpProd.AK118+$PanelProd.AK118+$SawnProd.AK118)^$E118)
#    Is equal to -57. The if condition should be on the whole expression result,
#    but it's only the sum of the secondary products production.


def consumption(ds: xarray.Dataset, t: int) -> xarray.DataArray:
    """Compute consumption equation 1"""
    return (
        ds["cons_constant"]
        * pow(ds["price"].loc[ds.c, t - 1], ds["cons_price_elasticity"])
        * pow(ds["gdp"].loc[ds.c, t], ds["cons_gdp_elasticity"])
    )


def consumption_pulp(
    ds: xarray.Dataset, ds_paper: xarray.Dataset, t: int
) -> xarray.DataArray:
    """Compute the Domestic Demand for Wood Pulp equation 2"""
    return (
        ds["cons_constant"]
        * pow(ds["price"].loc[ds.c, t - 1], ds["cons_price_elasticity"])
        * pow(
            ds_paper["prod"].loc[ds_paper.c, t], ds["cons_paper_production_elasticity"]
        )
    )


def consumption_indround(
    ds: xarray.Dataset,
    ds_sawn: xarray.Dataset,
    ds_panel: xarray.Dataset,
    ds_pulp: xarray.Dataset,
    t: int,
    compatible_mode: bool = None,
) -> xarray.DataArray:
    """Domestic demand for industrial roundwood equation 3

    The argument `compatible_mode` is to reproduce a behaviour in GFPMX-8-6-2021.xlsx
    for comparison purposes. Content of cell AJ2 in sheet IndroundCons:
    =IF($PulpProd.AJ2+$PanelProd.AJ2+$SawnProd.AJ2<=0,0,
        $G2*($IndroundPrice.AI2^$D2)*($PulpProd.AJ2+$PanelProd.AJ2+$SawnProd.AJ2)^$E2)
    The if condition is only the sum of the 3 secondary products production.
    Because Singapore has a negative constant (column G), it results in a negative consumption.
    """
    if compatible_mode is None:
        compatible_mode = GFPMX_8_6_2021_COMPATIBLE_MODE
    sum_prod_secondary = (
        ds_sawn["prod"].loc[ds_sawn.c, t]
        + ds_panel["prod"].loc[ds_panel.c, t]
        + ds_pulp["prod"].loc[ds_pulp.c, t]
    )
    cons = (
        ds["cons_constant"].loc[ds.c]
        * pow(
            ds["price"].loc[ds.c, t - 1],
            ds["cons_price_elasticity"].loc[ds.c],
        )
        * pow(
            sum_prod_secondary,
            ds["cons_products_elasticity"],
        )
    )
    if compatible_mode:
        # Keep only rows where sum_prod_secondary is positive
        cons.loc[sum_prod_secondary < 0] = 0
        return cons
    # Keep only rows where consumption is positive
    return np.maximum(cons, 0)


def import_demand(ds: xarray.Dataset, t: int) -> xarray.DataArray:
    """Compute import demand equation 4"""
    return (
        ds["imp_constant"]
        * pow(
            ds["price"].loc[ds.c, t - 1] * (1 + ds["tariff"].loc[:, t - 1]),
            ds["imp_price_elasticity"],
        )
        * pow(ds["gdp"].loc[ds.c, t], ds["imp_gdp_elasticity"])
    )


def import_demand_pulp(
    ds: xarray.Dataset, ds_paper: xarray.Dataset, t: int
) -> xarray.DataArray:
    """Compute the import demand for Wood Pulp equation 5"""
    # =$G2*(($PulpPrice.AI2*(1+$PulpTariff.AI2))^$D2)*($PaperProd.AJ2^$E2)
    return (
        ds["imp_constant"]
        * pow(
            ds["price"].loc[ds.c, t - 1] * (1 + ds["tariff"].loc[ds.c, t - 1]),
            ds["imp_price_elasticity"],
        )
        * pow(ds_paper["prod"].loc[ds.c, t], ds["imp_paper_production_elasticity"])
    )


def import_demand_indround(
    ds: xarray.Dataset,
    ds_sawn: xarray.Dataset,
    ds_panel: xarray.Dataset,
    ds_pulp: xarray.Dataset,
    t: int,
    compatible_mode: bool = None,
) -> xarray.DataArray:
    """Compute the import demand of industrial roundwood equation 6"""
    if compatible_mode is None:
        compatible_mode = GFPMX_8_6_2021_COMPATIBLE_MODE
    sum_prod_secondary = (
        ds_sawn["prod"].loc[ds_sawn.c, t]
        + ds_panel["prod"].loc[ds_panel.c, t]
        + ds_pulp["prod"].loc[ds_pulp.c, t]
    )
    imp = (
        ds["imp_constant"].loc[ds.c]
        * pow(
            (1 + ds["tariff"].loc[ds.c, t]) * ds["price"].loc[ds.c, t - 1],
            ds["imp_price_elasticity"].loc[ds.c],
        )
        * pow(sum_prod_secondary, ds["imp_products_elasticity"].loc[ds.c])
    )
    if compatible_mode:
        # Keep only rows where sum_prod_secondary is positive
        imp.loc[sum_prod_secondary < 0] = 0
        return imp
    # Keep only rows where import demand is positive
    return np.maximum(imp, 0)


def export_supply(ds: xarray.Dataset, t: int) -> xarray.DataArray:
    """Compute export supply equation 7

    Replace negative values by zero."""
    world_imp = ds["imp"].loc[ds.c, t].sum()
    exp = (
        ds["exp_marginal_propensity_to_export"].loc[ds.c] * world_imp
        + ds["exp_constant"]
    )
    return np.maximum(exp, 0)


def production(ds: xarray.Dataset, t: int) -> xarray.DataArray:
    """Compute domestic production equation 8
    Replace negative values by zero
    """
    prod = ds["cons"].loc[ds.c, t] + ds["exp"].loc[ds.c, t] - ds["imp"].loc[ds.c, t]
    return np.maximum(prod, 0)


def world_price(
    ds: xarray.Dataset, ds_primary: xarray.Dataset, t: int
) -> xarray.DataArray:
    """Compute the world price equation 10
    as a function of the input price
    """
    return ds["price_constant"].loc["WORLD"] * pow(
        ds_primary["price"].loc["WORLD", t], ds["price_input_elast"].loc["WORLD"]
    )


def world_price_indround(
    ds: xarray.Dataset, ds_other: xarray.Dataset, t: int
) -> xarray.DataArray:
    """Compute the world price of industrial roundwood equation 9"""
    # =$G182*($IndroundProd.AJ182^$F182)*($Stock.AJ182^$E182)*EXP($D182*AJ1)
    return (
        ds["price_constant"].loc["WORLD"]
        * pow(
            ds["prod"].loc["WORLD", t],
            ds["price_world_price_elasticity"].loc["WORLD"],
        )
        * pow(
            ds_other["stock"].loc["WORLD", t],
            ds["price_stock_elast"].loc["WORLD"],
        )
        * np.exp(ds["price_trend"].loc["WORLD"] * t)
    )


def local_price(ds: xarray.Dataset, t: int) -> xarray.DataArray:
    """Compute the local price equation 12"""
    return ds["price_constant"].loc[ds.c] * pow(
        ds["price"].loc["WORLD", t], ds["price_world_price_elasticity"].loc[ds.c]
    )


def forest_stock(
    ds: xarray.Dataset, ds_indround: xarray.Dataset, ds_fuel: xarray.Dataset, t: int
) -> xarray.DataArray:
    """Compute the forest stock expressed as growth drain equation 15

    Notes:
    - Roundwood is the sum of industrial round wood and fuel wood.
    - Replaces negative values by zero
    - Converts the stock from thousand m3 to million m3
        - TODO remove this conversion? Currently kept for comparison purposes with GFPMx.
            - Other units are in 1000 m3, should the whole model be harmonized to m3?
    """
    indround_fuel_prod = (
        ds_indround["prod"].loc[ds_indround.c, t - 1]
        + ds_fuel["prod"].loc[ds_fuel.c, t - 1]
    )
    stock = (
        ds["stock"].loc[ds.c, t - 1]
        * (1 + ds["stock_growth_rate_without_harvest"].loc[ds.c])
        - ds["stock_harvest_effect_on_stock"].loc[ds.c] * indround_fuel_prod / 1000
    )
    return np.maximum(stock, 0)


def compute_country_aggregates(
    ds: xarray.Dataset, t: int, variable: str = None
) -> None:
    """Compute aggregates for 'WORLD' and for
    'AFRICA', 'NORTH AMERICA', 'SOUTH AMERICA', 'ASIA', 'OCEANIA', 'EUROPE'
    param: ds dataset
    param: t time in years
    param: variable list of variables to aggregate in the dataset
    ! This function modifies its input data set `ds` for the given time step t.
    """
    regions = ds.region.to_series().unique()
    regions = [
        x for x in regions if x != "WORLD" and not pandas.isna(x) and not "" == x
    ]
    if variable is None:
        variable = ["cons", "exp", "imp", "prod"]
    if isinstance(variable, str):
        variable = [variable]
    for var in variable:
        if any(ds[var].loc[ds.c, t].isnull()):
            msg = f"NA values in {ds.product} {var}:\n"
            msg += f"{ds[var].loc[ds.c, t]}"
            raise ValueError(msg)
        ds[var].loc["WORLD", t] = ds[var].loc[ds.c, t].sum()
        v_agg = ds[var].loc[ds.c, t].groupby(ds["region"].loc[ds.c]).sum()
        # Rename the dimension from region to country to avoid a ValueError:
        #   new dimensions ('country',) must be a superset of existing
        #   dimensions ('region',)
        v_agg = v_agg.rename({"region": "country"})
        # Ensure v_agg has 'country' dimension coordinates in the same order as
        # the dataset slice on the left-hand side.
        v_agg_aligned = v_agg.reindex(country=ds[var].loc[regions, t].country)
        ds[var].loc[regions, t] = v_agg_aligned


def compute_secondary_product_ciep(
    ds: xarray.Dataset, ds_primary: xarray.Dataset, t: int
) -> None:
    """Compute consumption, import, export and production equations
    corresponding to a semi finished (secondary) product.

    ! This function modifies the input data set `ds` for the given time step t.
    """
    ds["cons"].loc[ds.c, t] = consumption(ds, t)
    ds["imp"].loc[ds.c, t] = import_demand(ds, t)
    ds["exp"].loc[ds.c, t] = export_supply(ds, t)
    ds["prod"].loc[ds.c, t] = production(ds, t)


def compute_secondary_product_price(
    ds: xarray.Dataset, ds_primary: xarray.Dataset, t: int
) -> None:
    """Compute world prices and local prices
    ! This function modifies the input data set `ds` for the given time step t.
    """
    ds["price"].loc["WORLD", t] = world_price(ds, ds_primary, t)
    ds["price"].loc[ds.c, t] = local_price(ds, t)


def compute_one_time_step(
    ds_indround, ds_fuel, ds_pulp, ds_sawn, ds_panel, ds_paper, ds_other, year
):
    """Modifies the input data sets in place
    TODO: change this to use the gfpmx_data as the unique argument
    This requires adding the boolean country indicator ds.c as a prerequisite"""
    # 1. Compute stock growth and drain from stock and production at t-1
    ds_other["stock"].loc[ds_other.c, year] = forest_stock(
        ds_other, ds_indround, ds_fuel, year
    )
    # 2. Compute cons, imp, exp and prod of secondary products
    compute_secondary_product_ciep(ds_sawn, ds_indround, year)
    compute_secondary_product_ciep(ds_fuel, ds_indround, year)
    compute_secondary_product_ciep(ds_panel, ds_indround, year)
    compute_secondary_product_ciep(ds_paper, ds_pulp, year)
    # 3. Compute cons, imp, exp and prod of primary products
    ds_pulp["cons"].loc[ds_pulp.c, year] = consumption_pulp(ds_pulp, ds_paper, year)
    ds_pulp["imp"].loc[ds_pulp.c, year] = import_demand_pulp(ds_pulp, ds_paper, year)
    ds_pulp["exp"].loc[ds_pulp.c, year] = export_supply(ds_pulp, year)
    ds_pulp["prod"].loc[ds_pulp.c, year] = production(ds_pulp, year)
    ds_indround["cons"].loc[ds_pulp.c, year] = consumption_indround(
        ds_indround, ds_sawn, ds_panel, ds_pulp, year
    )
    ds_indround["imp"].loc[ds_indround.c, year] = import_demand_indround(
        ds_indround, ds_sawn, ds_panel, ds_pulp, year
    )
    ds_indround["exp"].loc[ds_indround.c, year] = export_supply(ds_indround, year)
    ds_indround["prod"].loc[ds_indround.c, year] = production(ds_indround, year)
    # 4. Compute Country aggregates and prices
    compute_country_aggregates(ds_sawn, year)
    compute_country_aggregates(ds_panel, year)
    compute_country_aggregates(ds_pulp, year)
    compute_country_aggregates(ds_paper, year)
    compute_country_aggregates(ds_fuel, year)
    compute_country_aggregates(ds_indround, year)
    compute_country_aggregates(ds_other, year, "stock")
    # Compute the world price and the local price of indround
    ds_indround["price"].loc["WORLD", year] = world_price_indround(
        ds_indround, ds_other, year
    )
    ds_indround["price"].loc[ds_indround.c, year] = local_price(ds_indround, year)
    # The world price of indround is required to compute the price of secondary products
    assert not ds_indround["price"].loc["WORLD", year].isnull()
    compute_secondary_product_price(ds_sawn, ds_indround, year)
    compute_secondary_product_price(ds_fuel, ds_indround, year)
    compute_secondary_product_price(ds_panel, ds_indround, year)
    compute_secondary_product_price(ds_pulp, ds_indround, year)
    # The world price of ds_pulp is required to compute the price of ds_paper
    assert not ds_pulp["price"].loc["WORLD", year].isnull()
    compute_secondary_product_price(ds_paper, ds_pulp, year)



"""Explore the output of the Fair and SSP2 scenarios

Before you can explore scenario output, you should run the model first with
this script:

    scripts/running/run_fair_spp2.py

Then this script can load the output in NetCDF files to explore the scenario data.


See also:

- the `save_harvest_demand_to_eu_cbm_hat` function used to send cobwood output
  data to EU-CBM-HAT at the end of the running script.

- data preparation script for the CBM output data of the SSP2 FAIR paper
    ~/eu_cbm/eu_cbm_explore/scenarios/ssp2_fair_degrowth/ssp2_fair_owc_prepare_data.py


"""

import pandas
from cobwood.gfpmx import GFPMX

###########################################
# Load data without re-running the model. #
###########################################
gfpmxpikfair_fel1 = GFPMX(scenario="pikfair_fel1")
gfpmxpikssp2_fel1 = GFPMX(scenario="pikssp2_fel1")
gfpmxpikssp2 = GFPMX(scenario="pikssp2")

# Get irw and fw production data for the ssp2 and ssp2_fel1 scenarios
ssp2_fel1_prod = gfpmxpikssp2_fel1.get_df(product=["indround", "fuel"], var="prod")
ssp2_prod = gfpmxpikssp2.get_df(product=["indround", "fuel"], var="prod")
prod = pandas.concat([ssp2_fel1_prod, ssp2_prod]).reset_index()

# Create a multi-level column structure
prod_grouped = prod.set_index(["country", "year", "scenario"])[
    ["fuel_prod", "indround_prod"]
]
# Unstack the scenario level to create columns
prod_wide2 = prod_grouped.unstack("scenario")

# Convert from 1000 M3 to M3
print("Production units are expressed in", gfpmxpikssp2["indround"]["prod"].unit)
prod_wide2 = prod_wide2 * 1e3

# Flatten the column names to get format like 'fuel_prod_pikssp2'
prod_wide2.columns = [f"{col[0]}_{col[1]}" for col in prod_wide2.columns]
prod_wide2 = prod_wide2.reset_index()

# Display a data Array for countries only
fairsawn = gfpmxpikfair_fel1["sawn"]
print(fairsawn["cons"][fairsawn.c])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux.

This is a legacy script written before re-using the biotrade package.


JRC biomass Project.
Unit D1 Bioeconomy.

Download a compressed CSV file from the FAOSTAT API and save it in the data folder.

>>> from cobwood.faostat import faostat
>>> faostat.download("Forestry_E_All_Data_(Normalized).zip")

Read a compressed CSV

>>> from cobwood.faostat import faostat
>>> fo = faostat.read_csv("Forestry_E_All_Data_(Normalized).zip")
"""


# Third party modules
import re
import shutil
import urllib.request
from pathlib import Path
from zipfile import ZipFile
import pandas

# Internal modules
from cobwood import cobwood_data_dir

# Define a location to store data locally
faostat_data_folder = cobwood_data_dir / "faostat"
# Create the faostat data folder if it doesn't exist
if not Path(faostat_data_folder).exists():
    Path(faostat_data_folder).mkdir(parents=True)


class Faostat:
    """
    Download from the FAOSTAT API and store it.
    """

    # Define the location of the original source data
    url_base = "http://fenixservices.fao.org/faostat/static/bulkdownloads"
    # Define a location to store data locally
    data_folder = faostat_data_folder

    def download(self, file_name):
        """Download a compressed CSV file from the FAOSTAT API and save it in the data folder.

        Sample use

        >>> from cobwood.faostat import faostat
        >>> faostat.download("Forestry_E_All_Data_(Normalized).zip")
        """
        orig_url = self.url_base + "/" + file_name
        dest_file = self.data_folder + "/" + file_name
        # Define headers, otherwise faostat throws an Error 403: Forbidden
        # Based on https://stackoverflow.com/a/66591873/2641825
        header = {
            "User-Agent": "Mozilla/5.0 (X11; Linux x86_64) "
            "AppleWebKit/537.11 (KHTML, like Gecko) "
            "Chrome/23.0.1271.64 Safari/537.11",
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
            "Accept-Charset": "ISO-8859-1,utf-8;q=0.7,*;q=0.3",
            "Accept-Encoding": "none",
            "Accept-Language": "en-US,en;q=0.8",
            "Connection": "keep-alive",
        }
        # Request
        req = urllib.request.Request(url=orig_url, headers=header)
        with urllib.request.urlopen(req) as response:
            with open(dest_file, "wb") as f:
                shutil.copyfileobj(response, f)

    def read_csv(self, file_name):
        """Read a compressed CSV into a pandas data frame and rename its columns to snake case.

        Sample use

        >>> from cobwood.faostat import faostat
        >>> fo = faostat.read_csv("Forestry_E_All_Data_(Normalized).zip")
        """
        zip_file_name = self.data_folder + "/" + file_name
        # There are 2 csv files in the archive
        # Load the csv file with the same name as the archive itself
        csv_file_name = re.sub(".zip", ".csv", file_name)

        with ZipFile(zip_file_name).open(csv_file_name) as csv_file:
            df = pandas.read_csv(
                csv_file, sep=",", quotechar='"', encoding="ISO-8859-1"
            )
        # Rename columns to snake case
        df = df.rename(columns=lambda x: re.sub(r" ", "_", x).lower())
        return df


# Make a singleton #
faostat = Faostat()



"""The purpose of this script is to check data consistency.

For example the sum of country values should be equal to the world value.
"""
# TODO: compute world export/imports
# TODO: compute world production/consumption

import pandas
from numpy.testing import assert_allclose


def check_world_aggregates(df, df_agg, rtol=None):
    """Check that the world aggregate correspond to the sum of constituents
    Compare only columns where the sum makes sense

    :param df data frame: country data
    :param df_agg data frame: world and continent data
    :param rtol float: relative tolerance
    """
    if rtol is None:
        rtol = 1e-7
    # TODO: check that the continent aggregates match with
    # the sum of their constituents.
    cols_compare = [
        "cons",
        "cons_usd",
        "exp",
        "exp_usd",
        "imp",
        "imp_usd",
        "prod",
        "prod_usd",
        "gdp",
    ]
    idx = pandas.IndexSlice
    # TODO: keep only columns that are present
    # cols_compare = df[set(cols_compare) & set(df.columns)]
    world_sum_1 = df_agg.loc[idx[:, "WORLD"], cols_compare]
    world_sum_2 = df.groupby(["year"]).agg("sum")[cols_compare]
    assert_allclose(world_sum_1, world_sum_2, rtol=rtol)


def check_nrows_years_countries(df, dataset_name):
    """Check that the number of rows is equal to the number of years times the
    number of countries. Return a diagnostic message that can be printed
    """
    years = df.index.to_frame()["year"].unique()
    countries = df.index.to_frame()["country"].unique()
    assert len(years) * len(countries) == len(df)
    msg = "Number of years times the number of countries"
    msg += f" in the {dataset_name} data frame: "
    msg += f"{len(years)} * {len(countries)} = {len(years) * len(countries)}"
    return msg


##################################
# Post processing quality checks #
##################################
# TODO use np.testing.assert_allclose()
# Check that the computed values correspond to the original GFPMx values
# Compare only values after the base year


def compare_to_original_cobwood(df, variables=None, base_year=None):
    """Compare computed variables to the original GFTMx values
    After times t = base_year + 1
    """
    if variables is None:
        variables = ["cons", "imp", "exp", "prod", "price"]
    df_comp = df.query("year > @base_year + 1")
    for var in variables:
        try:
            assert_allclose(df_comp[var + "2"], df_comp[var], rtol=1e-6)
        except AssertionError as e:
            print(
                "The",
                var,
                "variable does not match with original GFPMx data:\n",
                str(e),
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
A script to compute GFPMX equations using xarray

Run this file with:

     ipython -i ~/repos/cobwood/scripts/experiment_with_xarray.py

The first attempt is the easiest one based on the long format. But since it is based on the long
format, we loose the units. It would be nice to have the units as dataset
attribute to each data array

    sawn_ds.attrs["units"] = "test"
    sawn_ds.price.attrs["units"] = "$/m3"

The third attempt distinguishes:

- two dimensional data such as production, consumption or prices
- one dimensional data such as price elasticities which are defined by country
- attributes such as units

"""

# Third party modules
import xarray

from cobwood.gfpmx_data import gfpmx_data

# Load data
indround_agg = gfpmx_data.get_agg_rows("indround")
sawn = gfpmx_data.get_country_rows("sawn", ["gdp"])
fuel = gfpmx_data.get_country_rows("fuel", ["gdp"])

##################################################################
# First attempt convert data frames into datasets or data arrays #
##################################################################

# "Convert a pandas.DataFrame into an xarray.Dataset
#
#    Each column will be converted into an independent variable in the
#    Dataset. If the dataframe's index is a MultiIndex, it will be expanded
#    into a tensor product of one-dimensional indices (filling in missing
#    values with NaN). This method will produce a Dataset very similar to
#    that on which the 'to_dataframe' method was called, except with
#    possibly redundant dimensions (since all dataset variables will have
#    the same dimensionality)
sawn_ds = xarray.Dataset.from_dataframe(sawn)
sawn_ds

# class DataArray
# to_pandas(self) -> 'DataArray | pd.Series | pd.DataFrame'
#     Convert this array into a pandas object with the same shape.
#     The type of the returned object depends on the number of DataArray
#     dimensions:
#     * 0D -> `xarray.DataArray`
#     * 1D -> `pandas.Series`
#     * 2D -> `pandas.DataFrame`
#     Only works for arrays with 2 or fewer dimensions.
#     The DataArray constructor performs the inverse transformation.
sawn_da_dumb = xarray.DataArray(sawn)
sawn_da_dumb

# Convert
# See help(sawn_ds.to_stacked_array)

# I don't manage to use to_stacked_array
# sawn_ds[["price", "prod", "faostat_name"]].to_stacked_array("variable","faostat_name")
# ValueError: All variables in the dataset must contain the dimensions ('year', 'country').

# Convert data set to array
sawn_da = sawn_ds[["price", "prod"]].to_array("variable")

index = ["year", "country", "faostat_name"]
sawn_ds2 = xarray.Dataset.from_dataframe(sawn.reset_index().set_index(index))
sawn_da2 = sawn_ds2[["price", "prod"]].to_array("variable")

# Display index values
sawn_ds2.indexes["year"]
sawn_ds2.indexes["country"]

index = ["year", "country", "faostat_name"]
fuel_ds2 = xarray.Dataset.from_dataframe(fuel.reset_index().set_index(index))
fuel_da2 = fuel_ds2[["price", "prod"]].to_array("variable")


########################################################
# Second attempt build data array from the wide format #
########################################################
sawn_price_wide = gfpmx_data.get_sheet_wide("sawnprice")
index = [
    "faostat_name",
    "element",
    "unit",
    "input_elast",
    "world_price_elasticity",
    "constant",
    "country",
]
sawn_price_da = xarray.DataArray(
    sawn_price_wide.set_index(index), dims=["dim_0", "year"]
)

# Data for the first year
sawn_price_da[:, 0]
sawn_price_da.loc[:, "value1992"]
# N'importe quoi
# sawn_price_da.to_dataset(dim="dim_0").to_dataframe()


####################################################################################
# Third attempt panel 2 dimensions, elasticities 1 dimension and unit as attribute #
####################################################################################
# TODO: place conversion functions in a module
# Rename this script to experiment_with_xarray.py
# Move the computation to compute_with_xarray.py
sawn = gfpmx_data.convert_sheets_to_dataset("sawn", ["gdp"])

# gdp = convert_to_2d_array(gfpmx_data.get_sheet_wide("gdp"))
# sawn["gdp"] = gdp

# Plot
continents = [
    "WORLD",
    "AFRICA",
    "NORTH AMERICA",
    "SOUTH AMERICA",
    "ASIA",
    "OCEANIA",
    "EUROPE",
]
# spda.loc[dict(country=continents)].plot(col="country", col_wrap=4)
# from matplotlib import pyplot as plt
# plt.show()


# # Compute demand
# t = 2019
#
# # pow(gdp.loc[:,t],)
#
# # Compute world price
#
# # Compute local price
# price_t = (spds["constant"]
#            * pow(spds["price"].loc[:,t-1], spds["elast"])
#           ) # TODO: add gdp elasticity


#############
# Questions #
#############
# How to put the year first, before the country?
# So it's compatible with compute_all_equations.py
# Maybe using assign_coords?
# https://docs.xarray.dev/en/stable/generated/xarray.DataArray.assign_coords.html#xarray.DataArray.assign_coords
# The fact that country is first, is because country are in rows and years in columns in the input data.
# Maybe this is better this way?
sawn.price.loc["Algeria", 1992]



"""The purpose of this script is to export data for JRC D4

Usage:

    ipython -i $HOME/repos/cobwood/scripts/export_jrc_d4.py

"""


from cobwood.gfpmx import GFPMX
from biotrade.faostat import faostat
import warnings

eu_countries = faostat.country_groups.eu_country_names
eu_countries += ["Netherlands"]

# Load output data, after a run has already been completed
gfpmx_pikssp2 = GFPMX(
    input_dir="gfpmx_base2021", base_year=2021, scenario="pikssp2_fel1"
)
gfpmx_pikfair = GFPMX(
    input_dir="gfpmx_base2021", base_year=2021, scenario="pikfair_fel1"
)

SELECTED_VARIABLES = [
    "cons",
    "prod",
    "exp",
    "imp",
    "price",
    "gdp",
]

print("Check units")
for var in ["cons", "prod", "price"]:
    print(var, gfpmx_pikssp2["paper"][var].unit)
msg = "There is an issue with the unit, it should be in tons instead of m3.\n"
msg += "see https://gitlab.com/bioeconomy/cobwood/cobwood/-/issues/11"
warnings.warn(msg)

##########################################
# Extract paper projections to csv files #
##########################################


def extract_for_d4(gfpmx_scenario, product):
    """Extract data for EU countries for d4"""
    df = (
        gfpmx_scenario["paper"][SELECTED_VARIABLES]
        .to_dataframe()
        .reset_index()
        .sort_values(["country", "year"])
    )
    selector = df["country"].isin(eu_countries)
    df["unit"] = "1000 t"
    df["price_unit"] = "$/t"
    df["gdp_unit"] = "1000 constant 2005 USD"
    df["product"] = product
    return df.loc[selector]


paper_ssp2 = extract_for_d4(gfpmx_pikssp2, "paper")
print("countries:\n", paper_ssp2["country"].unique())
paper_ssp2.to_csv("/tmp/paper_gfpmx_pik_ssp2_scenario.csv", index=False)

paper_fair = extract_for_d4(gfpmx_pikfair, "paper")
paper_fair.to_csv("/tmp/paper_gfpmx_pik_fair_scenario.csv", index=False)



""" Run the pik FAIR and ssp2 scenarios The purpose of this script is to use
the GDP Fair scenario of Bodirsky et al 2022 as an input to the GFPMx forest
sector model

Usage:

    ipython -i ~/repos/cobwood/scripts/running/run_fair_spp2.py

Dependency this script should be run to prepare the PIK GDP data:

    ipython -i ~/repos/cobwood/scripts/load_pik_data.py

Steps for the run:

1 Load GFPMX data
2 Load GDP Fair scenario data and readjust to the same GDP level as the
  historical series so that there is no discontinuity at the base year.
3 Create a GFPMX object with GFPMx data and GDP Fair data
4 Run the model with that GDP scenario
5 Save output to netcdf files. The GFPMX class should have a method that stores
  each dataset to a scenario output folder  :

    ds.to_netcdf("path/to/file.nc")

- Read the output later to plot the model results, starting with an overview of
  industrial roundwood and fuelwood harvest at EU level

    ds = xarray.open_dataset("path/to/file.nc")

Note: in CBM, the base year is the first year of the simulation, as illustrated
  by the condition `if self.year < self.country.base_year` (in
  `eu_cbm_hat/cbm/dynamic.py`) which governs the start of the harvest allocation
  tool.

"""

import pathlib
import pandas
import cobwood
from cobwood.gfpmx import GFPMX
from cobwood.gfpmx_data import convert_to_2d_array
from cobwood.gfpmx_equations import compute_country_aggregates
from biotrade.faostat import faostat  # only for an EU diagnostic plot

eu_cbm_data_dir = cobwood.data_dir.parent / "eu_cbm" / "eu_cbm_data"

##############################
# Create GFTMX model objects #
##############################
print("Create GFTMX model objects.")
msg = "Do you want to erase output and re-run the scenarios?"
if input(msg + "\nPlease confirm [y/n]:") != "y":
    raise ValueError("Cancelled.")

gfpmxb2021 = GFPMX(
    input_dir="gfpmx_base2021", base_year=2021, scenario="base_2021", rerun=True
)
# BAU SSP2 GDP projections from Bodirstky et al 2022
gfpmxpikssp2 = GFPMX(
    input_dir="gfpmx_base2021", base_year=2021, scenario="pikssp2", rerun=True
)
# FAIR GDP projections from Bodirstky et al 2022
gfpmxpikfair = GFPMX(
    input_dir="gfpmx_base2021", base_year=2021, scenario="pikfair", rerun=True
)
# Scenario for a fuel wood demand elasticity of 1 fel1
gfpmxpikfair_fel1 = GFPMX(
    input_dir="gfpmx_base2021", base_year=2021, scenario="pikfair_fel1", rerun=True
)
gfpmxpikssp2_fel1 = GFPMX(
    input_dir="gfpmx_base2021", base_year=2021, scenario="pikssp2_fel1", rerun=True
)


# Models that will be run here and transmitted for further processing to CBM
models = [
    gfpmxpikssp2,
    gfpmxpikfair,
    gfpmxpikssp2_fel1,
    gfpmxpikfair_fel1,
]


#######################################################
# Re-compute the aggregates for the historical period #
#######################################################
# There seems to be an issue in the GFPMX spreadsheet where some continents get
# inverted
print("Re-compute aggregates for the historical period.")
for model in models + [gfpmxb2021]:
    for this_product in model.products:
        for year in range(1995, 2022):
            compute_country_aggregates(model[this_product], year)
            compute_country_aggregates(model.other, year, ["area", "stock"])


#############################
# Load and prepare GDP data #
#############################
print("Prepare GDP data")
# The country comes from the gfpmx country name
# See the many merges in scripts/load_pik_data.py
gdp_comp = pandas.read_parquet(cobwood.data_dir / "pik" / "gdp_comp.parquet")


def get_gdp_wide(df: pandas.DataFrame, column_name: str, year_min: int = 1995):
    """Transform the given GDP column into wide format for transformation into
    a 2 dimensional data array"""
    index = ["country", "year"]
    return (
        df[index + [column_name]]
        .loc[df["year"] >= year_min]
        .assign(year=lambda x: "value_" + x["year"].astype(str))
        .pivot(index="country", columns="year", values=column_name)
        .reset_index()
    )


pik_fair = get_gdp_wide(gdp_comp, "pik_fair_adjgfpm2021")
# This is the place where we rename PIK BAU to PIK SSP2, because BAU has a
# different meaning in the forest dynamics scenarios (it means the business as
# usual in terms of other wood components removals)
pik_ssp2 = get_gdp_wide(gdp_comp, "pik_bau_adjgfpm2021")

# 3 different forms of GDP dataset inside the GFPMX object
# gfpmxb2021.data.get_sheet_wide("gdp")
# gfpmxb2021.data["gdp"]
# gfpmxb2021.gdp

# The GDP added before the run is self.gdp
# In the __init__ method, self.gdp is created from the sheet in wide format.
# self.gdp = convert_to_2d_array(self.data.get_sheet_wide("gdp")).

# Assign new GDP values to the GFTMX objects, reindex them like the existing gdp array
# so that they get empty values for the country aggregates
# Convert from million USD to 1000 USD
gfpmxpikssp2.gdp = convert_to_2d_array(pik_ssp2).reindex_like(gfpmxb2021.gdp) * 1e3
gfpmxpikfair.gdp = convert_to_2d_array(pik_fair).reindex_like(gfpmxb2021.gdp) * 1e3

# GDP diagnostic plot
df = pik_fair.set_index("country").transpose()
eu_countries = faostat.country_groups.eu_country_names
cols = [col for col in df.columns if col in eu_countries]
# df[cols].plot()
selector = df.index.str.contains("1990|2020|2050")
# Relative difference
df.loc[selector, cols].diff() / df.loc[selector, cols]
df.loc[selector, cols].sum(axis=1).diff() / df.loc[selector, cols].sum(axis=1)


# Issue with missing GDP
# selector = gfpmxpikfair.gdp.loc[:, 2022].isnull()
# print(gfpmxpikfair.gdp["country"][selector])

# selector = pik_fair["country"].str.contains("ina|uyana|ntill")
# pik_fair.loc[selector]
# selector = gdp_comp["country"].str.contains("ina|uyana|ntill")
# gdp_comp.loc[selector].query("year == 2020")
# gdp_comp.query("country_iso == 'CHN'")

# Set values of 'Netherlands Antilles (former)', 'French Guyana',
# To the same as the existing GDP projections in GFPMX 2021
selected_countries = ["Netherlands Antilles (former)", "French Guyana"]
gfpmxpikssp2.gdp.loc[selected_countries] = gfpmxb2021.gdp.loc[selected_countries]
gfpmxpikfair.gdp.loc[selected_countries] = gfpmxb2021.gdp.loc[selected_countries]

# Missing PIK GDP for China (CHN) and Netherlands (NLD)
# The ISO 3 codes are present in the PIK csv files, why do they get dropped?


# Assign fuelwood elasticities 1 scenarios the same GDP as the other fair and bau
gfpmxpikssp2_fel1.gdp = gfpmxpikssp2.gdp
gfpmxpikfair_fel1.gdp = gfpmxpikfair.gdp


########################################
# Change fuel wood demand elasticities #
########################################
# Change fuel wood demand elasticities to 1
gfpmxpikfair_fel1.fuel["cons_gdp_elasticity"].loc[gfpmxpikfair_fel1.fuel.c] = 1
gfpmxpikssp2_fel1.fuel["cons_gdp_elasticity"].loc[gfpmxpikssp2_fel1.fuel.c] = 1


def cons_constant(ds, t):
    """Compute back the constant given the historical consumption"""
    return ds["cons"].loc[ds.c, t] / (
        pow(ds["price"].loc[ds.c, t - 1], ds["cons_price_elasticity"])
        * pow(ds["gdp"].loc[ds.c, t], ds["cons_gdp_elasticity"])
    )


# Add GDP projections to secondary products datasets.
gfpmxpikfair_fel1.fuel["gdp"] = gfpmxpikfair_fel1.gdp
gfpmxpikssp2_fel1.fuel["gdp"] = gfpmxpikssp2_fel1.gdp

# Calibrate the constant so that it reproduces the value in 2021
gfpmxpikfair_fel1.fuel["cons_constant"].loc[gfpmxpikfair_fel1.fuel.c] = cons_constant(
    gfpmxpikfair_fel1.fuel, 2021
)
gfpmxpikssp2_fel1.fuel["cons_constant"].loc[gfpmxpikssp2_fel1.fuel.c] = cons_constant(
    gfpmxpikssp2_fel1.fuel, 2021
)

# gfpmxpikfair_fel1.fuel["cons_constant"].to_pandas().reset_index().query("country in ['Germany','France']")
# cons_constant(gfpmxpikfair_fel1.fuel,2021).to_pandas().reset_index().query("country in ['Germany','France']")


##################################################################
# Fix Czechia's high propensity to export due to Salvage logging #
##################################################################
# Between 2016 and 2021, there was a massive bark beetle attack followed by
# salvage logging that represents several times the yearly harvest and large
# amounts of exports. Looking at Czechia's consumption, production and trade of
# all products
#     >>> products = ["sawn", "panel", "fuel", "paper", "indround", "pulp"]
#     >>> ds_fair = [gfpmxpikfair[x] for x in products]
#     >>> for ds in df_fair:
#     >>>     facet_plot_by_var(ds, countries=["Czechia"])
# The scenario should reduce industrial roundwood and sawnwood exports for the period
# To a value that corresponds to the average 2015-2020
# This can be achieved by reducing the marginal propensity to export.


selected_scenarios = [gfpmxpikssp2_fel1, gfpmxpikfair_fel1]
# Estimates from cobwood/scripts/estimating/estimate_export_supply.py
for scenario in selected_scenarios:
    scenario["indround"]["exp_constant"].loc["Czechia"] = -395.63598204559366
    scenario["indround"]["exp_marginal_propensity_to_export"].loc[
        "Czechia"
    ] = 0.030772674789748013
    scenario["sawn"]["exp_constant"].loc["Czechia"] = 1344.7487441529158
    scenario["sawn"]["exp_marginal_propensity_to_export"].loc[
        "Czechia"
    ] = 0.0025397644268203446

# See plots in notebook


# # Check resulting export values
# from cobwood.gfpmx_equations import export_supply
# export_supply(gfpmxpikssp2_fel1["indround"], 2021).loc["Czechia"]
# export_supply(gfpmxpikssp2["indround"], 2021).loc["Czechia"]


# for scenario in selected_scenarios:
#     product = "indround"
#     file_name = f"/tmp/{scenario.scenario}_{product}.csv"
#     scenario[product].to_dataframe().loc[["WORLD", "Czechia"]].to_csv(file_name)


#######
# Run #
#######
for model in models + [gfpmxb2021]:
    model.run()


# Note: this loop could be vectorized on years to speed it up.


####################
# Save output data #
####################
# print("Save output data to CSV")
# Note: the model run instruction already saves the output in NetCDF files
# under cobwood_data/gfpmx_output
#

# Save output to csv files in cobwood_data
# print("Save output to csv")
# gfpmxpikfair.datasets
# A dictionary of datasets which we can use to loop or store results
# fair_dir = cobwood.create_data_dir("pikfair")
# for ds in [
#     gfpmxpikfair.indround,
#     gfpmxpikfair.sawn,
#     gfpmxpikfair.panel,
#     gfpmxpikfair.pulp,
#     gfpmxpikfair.paper,
#     gfpmxpikfair.fuel,
#     gfpmxpikfair.other,
# ]:
#     # If a data array is 2 dimensional, write it to disk
#     for this_var in ds.data_vars:
#         if ds[this_var].ndim == 2:
#             df = ds[this_var].to_pandas()
#             df.rename(columns=lambda x: "value_" + str(x), inplace=True)
#             df.reset_index(inplace=True)
#             file_name = ds.product + this_var + ".csv"
#             df.to_csv(fair_dir / file_name, index=False)
#             print(file_name, "\n", df.columns[[0, 1, -1]], "\n")


###################################
# Save output data to eu_cbm_data #
###################################
def da_to_csv(da, file_path, faostat_name):
    """Data set to eu_cbm_data csv file"""
    df = da.to_pandas()
    df.rename(columns=lambda x: "value_" + str(x), inplace=True)
    df.reset_index(inplace=True)
    # Add columns required by eu_cbm_hat
    df["faostat_name"] = faostat_name
    df["element"] = "Production"
    df["unit"] = "1000m3"
    # Place the last 3 columns first
    cols = list(df.columns)
    cols = cols[-3:] + cols[:-3]
    df = df[cols]
    # Write to CSV
    df.to_csv(file_path, index=False)
    print(f"Writing {df.shape[0]} rows and {df.shape[1]} columns to:")
    print("  ", file_path, "\n")


def save_harvest_demand_to_eu_cbm_hat(model):
    """Save harvest demand to eu_cbm_hat"""
    eu_cbm_harvest_dir = pathlib.Path(eu_cbm_data_dir) / "domestic_harvest"
    eu_cbm_harvest_dir = eu_cbm_harvest_dir / model.scenario
    eu_cbm_harvest_dir.mkdir(exist_ok=True)
    da_to_csv(
        model.indround["prod"],
        eu_cbm_harvest_dir / "irw_harvest.csv",
        "Industrial roundwood",
    )
    da_to_csv(model.fuel["prod"], eu_cbm_harvest_dir / "fw_harvest.csv", "Fuelwood")


for model in models:
    save_harvest_demand_to_eu_cbm_hat(model)



"""Compare output where the fuel wood elasticity was set to one to the base model

Run this script

    penv
    ipython -i /home/paul/rp/cobwood/scripts/compare_fel1_to_base.py

"""

import sys
import pandas
import matplotlib.pyplot as plt
from biotrade.faostat import faostat
import cobwood
from cobwood.gfpmx import GFPMX

eu_countries = faostat.country_groups.eu_country_names
eu_countries += ["Netherlands"]

# Load CBM output data
eu_cbm_explore_path = cobwood.data_dir.parent / "eu_cbm" / "eu_cbm_explore"
scenario_dir = eu_cbm_explore_path / "scenarios" / "ssp2_fair_degrowth"
# Load data and plotting functions from the prepare script in the same directory
sys.path.append(str(scenario_dir))


hwp_dir = cobwood.data_dir / "gfpmx_output" / "hwp"
hwp_dir.mkdir(exist_ok=True)

eu_countries = faostat.country_groups.eu_country_names
eu_countries += ["Netherlands"]

# Load output data, after a run has already been completed
gfpmx_pikssp2_fel1 = GFPMX(
    input_dir="gfpmx_base2021", base_year=2021, scenario="pikssp2_fel1"
)
gfpmx_pikssp2_base = GFPMX(
    input_dir="gfpmx_base2021", base_year=2021, scenario="pikssp2"
)
gfpmx_pikfair_fel1 = GFPMX(
    input_dir="gfpmx_base2021", base_year=2021, scenario="pikfair_fel1"
)
gfpmx_pikfair_base = GFPMX(
    input_dir="gfpmx_base2021", base_year=2021, scenario="pikfair"
)


def get_df_eu(model, product, variable):
    """Extract a data frame from the model xarray output keep only EU countries"""
    df = model[product][variable].to_dataframe().reset_index()
    selector = df["country"].isin(eu_countries)
    df = df.melt(id_vars=["country", "year"], value_vars="cons", var_name="variable")
    df["product"] = product
    df["scenario"] = model.scenario
    return df.loc[selector].reset_index(drop=True)


########################################################
# Comparison of EU fuelwood consumption and production #
########################################################
comp_list = []
for this_prod in ["fuel", "indround"]:
    for this_var in ["cons", "prod"]:
        dfssp2base = get_df_eu(gfpmx_pikssp2_base, this_prod, this_var)
        dfssp2fel1 = get_df_eu(gfpmx_pikssp2_fel1, this_prod, this_var)
        dffairbase = get_df_eu(gfpmx_pikfair_base, this_prod, this_var)
        dffairfel1 = get_df_eu(gfpmx_pikfair_fel1, this_prod, this_var)
        comp_list += [dfssp2base, dfssp2fel1, dffairbase, dffairfel1]

comp = pandas.concat(comp_list)
del comp_list
index = ["year", "product", "scenario", "variable"]
compeu = comp.groupby(index)["value"].agg("sum").reset_index()


########
# Plot #
########
# Facet plot by country with 2 colors for the different sources
def plotcomp(df, product, variable):
    selector = df["product"] == product
    selector &= df["variable"] == variable
    selector &= df["year"] != 2071
    df = df.loc[selector].copy()
    df["value"] *= 1e-6
    df_wide = df.pivot(columns="scenario", index="year", values="value")
    df_wide.plot(title=f"{product} {variable} Million m3")
    plt.show()


plotcomp(compeu, "fuel", "cons")

plotcomp(compeu, "fuel", "prod")
plotcomp(compeu, "indround", "prod")
plotcomp(compeu, "indround", "cons")

# Percentage difference between base and fel1
# In 2030, 2050 and 2070
compeu["base_or_fel1"] = "base"
compeu[["scenario2", "model_version"]] = compeu["scenario"].str.split(
    "_", n=1, expand=True
)
selector = compeu["scenario"].str.contains("_fel1")
compeu.loc[~selector, "model_version"] = "base"

index = ["year", "product", "variable", "scenario2"]
compeu_wide = compeu.pivot(columns="base_or_fel1", index=index, values="value")
compeu_wide["reldiff"] = compeu_wide["fel1"] / compeu_wide["base"] - 1
compeu_wide.query("year in [2030, 2050]")

# Roundwood = indround + Fuelwood 2050
#                                   base	fel1	reldiff
# indround	cons	pikfair	471085.040325443	415235.488134672	-0.118555138478157
# indround	prod	pikssp2	633721.456237559	649287.283194183	0.0245625689384723



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
A script to save all sheets of a GFPMX Excel files into csv files.


Usage example on all available GFPMX spreadsheets:

    >>> from cobwood.gfpmx_spreadsheet_to_csv import gfpmx_spreadsheet_to_csv
    >>> gfpmx_spreadsheet_to_csv("~/large_models/GFPMX-8-6-2021.xlsx")
    >>> gfpmx_spreadsheet_to_csv("~/large_models/GFPMX-base2020.xlsx")
    >>> gfpmx_spreadsheet_to_csv("~/large_models/GFPMX-base2021.xlsb")

Note the last file is an `xlsb` in binary format. You might get an ImportError:
Missing optional dependency 'pyxlsb'. Use pip or conda to install pyxlsb.

It can also be used as a script directly to export the default spreadsheet to csv:

    ipython -i ~/repos/cobwood/cobwood/gfpmx_spreadsheet_to_csv.py

Source of the GFPMX spread sheet data by Joseph Buongiorno:
    https://buongiorno.russell.wisc.edu/gfpm/

"""

# Built-in modules #
from pathlib import Path
import re

# Third party modules #
import pandas
from tqdm import tqdm

# First party modules #

# Internal modules #
import cobwood


def extract_world_price_parameter(df, col_name, contains, var_name):
    """Extract world price parameters such as input elasticity,
    stock elasticity and trend.
    """
    selector = df[col_name].astype(str).str.contains(contains)
    if any(selector):
        df = df.rename(columns={col_name: var_name})
        # Force variable type to numeric
        df[var_name] = pandas.to_numeric(df[var_name], errors="coerce")
    return df


def gfpmx_spreadsheet_to_csv(spreadsheet_path):
    """Convert GFTMX data to csv files and store them inside the cobwood_data directory

    The function gfpmx_spreadsheet_to_csv loads a Excel file and writes all
    sheets into individual CSV files. The CSV files are saved inside
    gfpmx_data_dir under a subdirectory name that correspond to the input file
    name converted to lower case. Different calls to this function can load 2
    or more different input files  based on different input Excel files.

    Usage:

        >>> from cobwood.gfpmx_spreadsheet_to_csv import gfpmx_spreadsheet_to_csv
        >>> gfpmx_spreadsheet_to_csv("~/large_models/GFPMX-8-6-2021.xlsx")
        >>> gfpmx_spreadsheet_to_csv("~/large_models/GFPMX-base2020.xlsx")
        >>> gfpmx_spreadsheet_to_csv("~/large_models/GFPMX-base2021.xlsb")

    """
    # Convert the input file name to snake case. Use this as the output directory name.
    gfpmx_data_dir = re.sub(r"\W+", "_", Path(spreadsheet_path).stem).lower()
    gfpmx_data_dir = cobwood.data_dir / "gfpmx_input" / gfpmx_data_dir
    if not gfpmx_data_dir.exists():
        gfpmx_data_dir.mkdir(parents=True)
    print(f"\nLoad the Excel file {spreadsheet_path} in a dictionary of data frames.")
    gfpmx_excel_file = pandas.read_excel(spreadsheet_path, sheet_name=None)
    print(gfpmx_excel_file.keys())
    print(f"\nWrite all sheets to csv files in the output folder\n{gfpmx_data_dir}")
    for key in tqdm(gfpmx_excel_file.keys()):
        df = gfpmx_excel_file[key]
        # Remove empty columns in place
        df.dropna(how="all", axis=1, inplace=True)
        # Rename columns to snake case, replace all non alphanumeric characters by an underscore
        df.rename(columns=lambda x: re.sub(r"\W+", "_", str(x)).lower(), inplace=True)
        # Add "value" prefix to year columns in preparation for a reshape from wide to long
        df.rename(columns=lambda x: re.sub(r"^(\d{4})$", r"value_\1", x), inplace=True)
        # Rename element and unit columns
        df.rename(columns={"unnamed_1": "element", "unnamed_2": "unit"}, inplace=True)
        # Rename the column that contains the world price elasticity of the input
        # Equation 9 World Price of Industrial Roundwood
        # Equation 10 industrial roundwood world price elasticity
        # Equation 11 pulp world price elasticity
        price_tables = [
            "FuelPrice",
            "SawnPrice",
            "PanelPrice",
            "PulpPrice",
        ]
        if key in price_tables:
            df = extract_world_price_parameter(df, "unnamed_4", "ound", "input_elast")
        if key == "PaperPrice":
            df = extract_world_price_parameter(df, "unnamed_4", "ulp", "input_elast")
        if key == "IndroundPrice":
            df = extract_world_price_parameter(df, "unnamed_3", "trend", "trend")
            df = extract_world_price_parameter(df, "unnamed_4", "stock", "stock_elast")
        if key == "IndroundExp":
            if "marginal_propensity_to_export" not in df.columns:
                df.rename(
                    columns={"unnamed_5": "marginal_propensity_to_export"}, inplace=True
                )
            if "constant" not in df.columns:
                df.rename(columns={"unnamed_6": "constant"}, inplace=True)

        # This applies to product sheets, which have a "faostat_name" column
        if "faostat_name" in df.columns:
            # Harmonize product names
            if df["faostat_name"].unique().tolist() == ["Sawnwood"]:
                df["faostat_name"] = "Sawnwood+sleepers"

            # Fix typo in roundwood name
            df["faostat_name"].replace({"Roundwwood": "Roundwood"}, inplace=True)

            # Remove rows that don't have a faostat_name
            # They usually contain quality checks such as
            # World prod/cons or Worldexp/Worldimp
            df = df[~df["faostat_name"].isna()].copy()

        # Harmonize column names
        if "world_elasticity" in df.columns:
            df.rename(
                columns={"world_elasticity": "world_price_elasticity"}, inplace=True
            )
        if "stock_growth_rate_without_harvest" in df.columns:
            df.rename(
                columns={
                    "stock_growth_rate_without_harvest": "growth_rate_without_harvest"
                },
                inplace=True,
            )
        # Capitalize WORLD
        # It's written "World" in "FuelPrice" GFPMX base 2021
        if "country" in df.columns:
            df["country"] = df["country"].replace("World", "WORLD")

        # Further renaming for the purpose of libcbm usage
        if key in ["FuelProd", "IndroundProd"]:
            df.faostat_name = df.faostat_name.ffill()
            df.element = df.element.ffill()
            df.unit = df.unit.ffill()
        # Write the csv file
        csv_file_name = re.sub(r"\$", r"_usd", key).lower() + ".csv"
        csv_file_name = Path(gfpmx_data_dir) / csv_file_name
        df.to_csv(csv_file_name, index=False)


if __name__ == "__main__":
    # Input file from https://buongiorno.russell.wisc.edu/gfpm/
    gfpmx_spreadsheet_to_csv("~/large_models/GFPMX-8-6-2021.xlsx")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Give access to the dataset from the GFPMX model by Joseph Buongiorno.
Originally released at
    https://buongiorno.russell.wisc.edu/gfpm/

Before using this object, the Excel file needs to be exported to csv files with:

    >>> from cobwood.gfpmx_spreadsheet_to_csv import gfpmx_spreadsheet_to_csv
    >>> gfpmx_spreadsheet_to_csv("~/large_models/GFPMX-8-6-2021.xlsx")
    >>> gfpmx_spreadsheet_to_csv("~/large_models/GFPMX-base2020.xlsx")
    >>> gfpmx_spreadsheet_to_csv("~/large_models/GFPMX-base2021.xlsb")

The data will then be available in a sub directory of `cobweb.input_dir` with
the same name as the spreadsheet file (except that it will be in snake case
`gfpmx_8_6_2021`).

# TODO: update examples to use GFPMX.input_data instead of using this directly

    >>> from cobwood.gfpmx_data import GFPMXData
    >>> gfpmx_data_b2018 = GFPMXData(input_dir="gfpmx_8_6_2021")
    >>> gfpmx_data_b2020 = GFPMXData(input_dir="gfpmx_base2020")
    >>> gfpmx_data_b2021 = GFPMXData(input_dir="gfpmx_base2021")

You can view spreadsheets contents (loaded from intermediate csv files) by
selecting their names:

    >>> gfpmx_data_b2018.sheets
    >>> gfpmx_data_b2018["indroundprod"]
    >>> gfpmx_data_b2018["stock"]

You can load Xarray datasets with the `convert_sheets_to_dataset()` method.

    >>> indround = gfpmx_data_b2018.convert_sheets_to_dataset("indround")
    >>> sawn = gfpmx_data_b2018.convert_sheets_to_dataset("sawn")
    >>> other = gfpmx_data_b2018.convert_sheets_to_dataset("other")

"""

# Built in modules
import re

# Third party modules
import numpy as np
import pandas
import xarray
from numpy.testing import assert_allclose


def convert_to_2d_array(df: pandas.DataFrame) -> xarray.DataArray:
    """Convert the year columns of a data frame to a two dimensional data array
    In wide format the values are in one column for each year.

    Example use:

        >>> from cobwood.gfpmx_data import GFPMXData
        >>> from cobwood.gfpmx_data import convert_to_2d_array
        >>> gfpmx_data = GFPMXData(input_dir="gfpmx_8_6_2021")
        >>> sawnprice_df = gfpmx_data.get_sheet_wide("sawnprice")
        >>> sawnprice_da = convert_to_2d_array(sawnprice_df)
        >>> gdp_df = gfpmx_data.get_sheet_wide("gdp")
        >>> gdp_da = convert_to_2d_array(gdp_df)

    """
    cols = df.columns
    value_cols = cols[df.columns.str.contains("value")]
    da = xarray.DataArray(df.set_index("country")[value_cols], dims=["country", "year"])
    # Change year to an integer
    da["year"] = da["year"].str.replace("value_", "").astype(int)
    if "unit" in df.columns:
        # Store the unit as an attribute
        da.attrs["unit"] = df["unit"].unique()[0]
    return da


def convert_to_1d_array(df: pandas.DataFrame, var: str) -> xarray.DataArray:
    """Convert one column of a data frame to a one dimensional data array

    Example use:

        >>> from cobwood.gfpmx_data import GFPMXData
        >>> from cobwood.gfpmx_data import convert_to_1d_array
        >>> gfpmx_data = GFPMXData(input_dir="gfpmx_8_6_2021")
        >>> sawnprice_df = gfpmx_data.get_sheet_wide("sawnprice")
        >>> sawnprice_elast_da = convert_to_1d_array(sawnprice_df, "world_price_elasticity")

    """
    return xarray.DataArray(df.set_index("country")[var], dims=["country"])


def check_variable_presence(ds: xarray.Dataset) -> None:
    """Check that the required variables and coefficients are present to
    compute equations. Since various products have different equations, the
    requirements differ among primary and secondary products.
    """
    common_vars = [
        "cons",
        "cons_constant",
        "cons_price_elasticity",
        "cons_usd",
        "exp",
        "exp_usd",
        "exp_constant",
        "exp_marginal_propensity_to_export",
        "imp",
        "imp_constant",
        "imp_price_elasticity",
        "imp_usd",
        "nettrade",
        "nettrade_usd",
        "price",
        "price_constant",
        "price_world_price_elasticity",
        "prod",
        "prod_usd",
        "tariff",
    ]
    indround_vars = [
        "cons_products_elasticity",
        "imp_products_elasticity",
        "price_stock_elast",
    ]
    end_product_vars = [
        "cons_gdp_elasticity",
        "conspercap",
        "imp_gdp_elasticity",
        "price_input_elast",
    ]
    other_vars = [
        "area",
        "area_constant",
        "gdp",
        "gdppercap",
        "harvestperha",
        "harvestperstock",
        "population",
        "stock",
        "stock_harvest_effect_on_stock",
        "stock_growth_rate_without_harvest",
        "stockpercap",
        "stockperha",
        "totalcons_usd",
        "totalexp_usd",
        "totalimp_usd",
        "totalnettrade_usd",
        "totalprod_usd",
        "valueadded",
    ]
    # TODO: add checks for fuel wood
    data_vars = list(ds.data_vars)
    required_vars = set()
    if ds.attrs["product"] == "pulp":
        required_vars = common_vars
    if ds.attrs["product"] == "indround":
        required_vars = common_vars + indround_vars
    if ds.attrs["product"] in ["sawn", "panel", "paper"]:
        required_vars = common_vars + end_product_vars
    if ds.attrs["product"] == "other":
        required_vars = other_vars
    # Check that the required variables are present
    if not set(required_vars).issubset(data_vars):
        msg = "The following variables are missing from "
        msg += f"{ds.attrs['product']}:\n"
        msg += f"{set(required_vars) - set(data_vars)}"
        raise ValueError(msg)


def remove_after_base_year_and_copy(ds: xarray.Dataset, base_year):
    """Remove values after the base year and return a deep copy of the
    input dataset, i.e. the input dataset is not modified.
    This applies only to arrays which have a time dimension.
    Keep future tariff data, since they are exogenous.
    """
    ds_out = ds.copy(deep=True)
    for x in ds_out.data_vars:
        if len(ds_out[x].dims) == 2 and "tariff" not in x:
            ds_out[x].loc[dict(year=ds_out.coords["year"] > base_year)] = np.nan
    return ds_out


def compare_to_ref(
    ds: xarray.Dataset,
    ds_ref: xarray.Dataset,
    variable: [list, str],
    t: int,
    rtol: int = None,
    strict: bool = True,
):
    """Compare the computed dataset to the reference dataset for the given t
    Example use:
        >>> compare_to_ref(sawn, sawn_ref, "price", 2019)
        >>> compare_to_ref(indround, indround_ref, ["cons", "imp"], 2019)
    """
    if rtol is None:
        rtol = 1e-6
    if isinstance(variable, str):
        variable = [variable]
    final_message = "."
    for var in variable:
        # Production requires a different tolerance for some reason
        if var == "prod":
            rtol = 1e-2
        try:
            assert_allclose(
                ds[var].loc[ds.c, t],
                ds_ref[var].loc[ds.c, t],
                rtol=rtol,
            )
        except AssertionError as e:
            first_line_of_error = "".join(str(e).split("\n")[:3])
            msg = f"{ds.product}, {var}, {t}: {first_line_of_error}"
            if strict:
                raise AssertionError(msg) from e
            else:
                final_message = "There were errors."
                print(e, msg)
    print(f"Check {ds.product} {', '.join(variable)}: {final_message}")


class GFPMXData:
    """
    Read data from the GFTMX data set.

    The GFTMX dataset was converted to csv files one for each sheet in the
    original Excel Spreadsheet. This singleton gives access to each file.

    Load sawnwood consumption data in long format:

    :param input_dir Location of the csv files
    :param base_year Simulation base year i.e. last year of historical data
           available in the spreadsheet

        >>> from cobwood.gfpmx_data import GFPMXData
        >>> gfpmx_data = GFPMXData(input_dir="gfpmx_8_6_2021")
        >>> swd_cons = gfpmx_data['sawncons']
        >>> swd_cons

    The GFPMX dataset is useful to:

        1. Verify the reproducibility of results given in the spreadsheet

        2. Reuse elasticities, tariffs, constants and other coefficients that cannot be
           estimated easily

    See also the script that moves data from the original Excel spreadsheet to CSV files:
    `scripts/gfpmx_spreadsheet_to_csv.py`
    """

    def __getitem__(self, sheet_name):
        """Return a data frame based on the GFPMX sheet name."""
        return self.get_sheet_long(sheet_name)

    def __init__(self, parent):
        self.parent = parent
        self.input_dir = self.parent.input_dir
        if not self.input_dir.exists():
            msg = "The input data directory specified in the scenario yaml file "
            msg += "doesn't exist.\n"
            msg += f"Input data directory: {self.input_dir}\n"
            msg += "Configuration file:"
            msg += f"{self.parent.scenario_yaml_path}\n"
            msg += f"{self.parent.config}\n"
            raise FileNotFoundError(msg)
        self.sheets = self.list_sheets()
        self.index_merge = ["year", "country", "faostat_name"]
        self.index = ["year", "country"]
        self.country_groups = self.get_country_groups()
        self.country_aggregates = [
            "WORLD",
            "AFRICA",
            "NORTH AMERICA",
            "SOUTH AMERICA",
            "ASIA",
            "OCEANIA",
            "EUROPE",
        ]
        assert set(self.country_aggregates) - set(self.country_groups["region"]) == {
            "WORLD"
        }

    def list_sheets(self):
        """List sheets available in the GFPMX data folder

        :return data frame with the sheet name, product and element

        For example show all sheets available

            >>> from cobwood.gfpmx_data import GFPMXData
            >>> from pandas.errors import EmptyDataError
            # TODO: update examples to use GFPMX.input_data
            >>> gfpmx_data = GFPMXData(input_dir="gfpmx_8_6_2021")
            >>> sheets = gfpmx_data.list_sheets()
            >>> sheets

        As a prerequisite to merge sheets together, the following code
        shows additional variables besides the value in each sheet:

            >>> known_columns = ['year', 'element', 'unit', 'country',
            >>>                  'faostat_name', 'value']
            >>> for prod in sheets["product"].unique():
            >>>     sheets_selected = sheets.query("product==@prod")
            >>>     print(f"Additional variables in '{prod}' related sheets:")
            >>>     for s in sheets_selected.index:
            >>>         try:
            >>>             df = gfpmx_data[s]
            >>>             columns = df.columns
            >>>         except EmptyDataError:
            >>>             print(f"   No data in the '{s}' file.")
            >>>             columns = ["no data"]
            >>>         print("  ", s, set(columns) - set(known_columns))

        List all columns in the roundwood related sheets

            >>> for name in sheets.query("product == 'round'").index:
            >>>     print(name, "\n",  gfpmx_data[name].columns.tolist())

        List all columns in the other sheets

            >>> for name in sheets.query("product == 'other'").index:
            >>>     if name in ["author", "names", "notes", "worldprice"]:
            >>>          continue
            >>>     print(name, "\n",  gfpmx_data[name].columns.tolist())

        Display the shape of the other sheets

            >>> for s in sheets.query("product == 'other'").index:
            >>>     try:
            >>>         print(s, gfpmx_data[s].shape)
            >>>     except EmptyDataError:
            >>>         print(f"No data in the '{s}' file.")
            >>>     except ValueError as e:
            >>>         print(f"Error in {s}: {e}")

        Display the faostat_name of the sawnwood sheets

            >>> for name in sheets.query("product == 'sawn'").index:
            >>>     print(name, "\n",  gfpmx_data[name]["faostat_name"].unique())

        """
        sheet_paths = self.input_dir.glob("**/*.csv")
        df = pandas.DataFrame({"file_name": [x.name for x in sheet_paths]})
        df["name"] = df.file_name.str.extract("(.*).csv")
        # Place product patterns in a capture group for extraction
        product_pattern = "fuel|indround|panel|paper|pulp|round|sawn"
        # Create a product column and an element column
        df[["product", "element"]] = df.name.str.extract(f"({product_pattern})?(.*)")
        df = df.sort_values(by=["product", "element"])
        df["product"] = df["product"].fillna("other")
        df = df[["name", "product", "element"]]
        df.set_index("name", inplace=True)
        return df

    def get_sheet_wide(self, sheet_name):
        """Read a csv file into a pandas data frame

        Example use

            >>> from cobwood.gfpmx_data import GFPMXData
            >>> gfpmx_data = GFPMXData(input_dir="gfpmx_8_6_2021")
            >>> print(gfpmx_data.get_sheet_wide("sawnprice"))

        """
        csv_file_name = self.input_dir / (sheet_name + ".csv")
        df = pandas.read_csv(csv_file_name)
        return df

    def get_sheet_long(self, sheet_name):
        """Read a csv file into a pandas data frame and reshape it to long format

        Example use

            >>> from cobwood.gfpmx_data import GFPMXData
            >>> gfpmx_data = GFPMXData(input_dir="gfpmx_8_6_2021")
            >>> print(gfpmx_data.get_sheet_long("sawncons"))

        """
        df_wide = self.get_sheet_wide(sheet_name=sheet_name)
        # Remove the underscore otherwise pandas.wide_to_long doesn't work
        df_wide.rename(
            columns=lambda x: re.sub(r"value_", "value", str(x)), inplace=True
        )
        # Reshape year columns to long format
        index = [x for x in df_wide.columns if not re.search("value", x)]
        df = pandas.wide_to_long(df_wide, stubnames="value", i=index, j="year")
        df.reset_index(inplace=True)
        # Rename the value column according to the shorter element part of the
        # file name. Note there is also an element column which we don't use
        # here.
        element = self.sheets.loc[sheet_name, "element"]
        df.rename(columns={"value": element}, inplace=True)

        # Prefix any columns that are not part of the index,
        # with the short element name
        index = ["faostat_name", "element", "unit", "country", "year", element]
        other_cols = list(set(df.columns) - set(index))
        other_cols_renamed = [element + "_" + x for x in other_cols]
        mapping = dict(zip(other_cols, other_cols_renamed))
        df.rename(columns=mapping, inplace=True)

        # Check that years are complete
        years = df["year"].unique()
        if not years.min() + len(years) - 1 == years.max():
            msg = f"The time series of '{sheet_name}' is not complete. "
            msg += "The following years are missing:\n"
            msg += str(set(range(years.min(), years.max() + 1)) - set(years))
            raise ValueError(msg)

        return df

    def get_gdp(self, sheet_name="gdp", index=None, var_name="gdp"):
        """Return a data frame of cleaned GDP values

        >>> from cobwood.gfpmx_data import GFPMXData
        >>> gfpmx_data = GFPMXData(input_dir="gfpmx_8_6_2021")
        >>> gfpmx_data.get_price_lag('sawnprice')

        """
        if index is None:
            index = ["id", "year", "country"]
        df = self.get_sheet_long(sheet_name)
        df = df[index + ["value"]]
        df.rename(columns={"value": var_name}, inplace=True)
        return df

    def get_price_lag(self, sheet_name, index=None, var_name="price"):
        """Return a price table with prices shifted by a one year lag

        >>> from cobwood.gfpmx_data import GFPMXData
        >>> gfpmx_data = GFPMXData(input_dir="gfpmx_8_6_2021")
        >>> gfpmx_data.get_price_lag('sawnprice')

        """
        if index is None:
            index = ["id", "year", "country"]
        df = self.get_sheet_long(sheet_name)
        df = df[index + ["value"]]
        df.rename(columns={"value": var_name}, inplace=True)
        # Shift prices by a one year lag
        df.set_index("year", inplace=True)
        df[var_name + "_lag"] = df.groupby(["id", "country"])[var_name].shift()
        df.reset_index(inplace=True)
        return df

    def join_sheets(self, product, other_element=None):
        """
        Merge all related data frames for a given product

        :param str product: selected product
        :param list other_element: list of element to be added from the
        rows where product is equal to "other"
        :return data frame

        For example join all roundwood sheets in one data frame and add a
        stock column.

            >>> from cobwood.gfpmx_data import GFPMXData
            >>> gfpmx_data = GFPMXData(input_dir="gfpmx_8_6_2021")
            >>> rwd = gfpmx_data.join_sheets("round", ["stock"])
            >>> rwd.columns

        Join sheets for all available products

            >>> for product in gfpmx_data.sheets["product"].unique():
            >>>     if product == "other":
            >>>         continue
            >>>     print(product)
            >>>     print(gfpmx_data.join_sheets(product).head())

        Join all other sheets

            >>> sheets = gfpmx_data.list_sheets()
            >>> other_sheets = sheets.query("product == 'other'")
            >>> other_element = other_sheets["element"].to_list()
            >>> for element in ["author", "names", "notes", "worldprice"]:
            >>>     other_element.remove(element)
            >>> other = gfpmx_data.join_sheets("round", other_element)
            >>> print(other.columns)
            >>> other[other.columns[other.columns.str.contains("unnamed")]]

        """

        # Join the product sheets together
        sheets = self.sheets[self.sheets["product"] == product]
        first_sheet = sheets.index[0]
        df_all = self.get_sheet_long(first_sheet)
        for name in sheets.index[1:]:
            df = self.get_sheet_long(name)
            # Keep only index columns or columns starting with element
            element = sheets.loc[name]["element"]
            cols = df.columns[df.columns.str.match("^" + element)].tolist()
            df = df[self.index_merge + cols]
            df_all = df_all.merge(df, "left", self.index_merge)
            if df_all[cols].sum(numeric_only=True).sum() == 0:
                raise ValueError(
                    "No data in %s \n %s \n %s" % (cols, df.head(), df_all.head())
                )

        # Join other sheets if requested
        if other_element is not None:
            other_sheets = self.sheets[self.sheets["product"] == "other"]
            other_sheets = other_sheets[other_sheets["element"].isin(other_element)]
            for name in other_sheets.index:
                df = self.get_sheet_long(name)
                element = other_sheets.loc[name]["element"]
                cols = df.columns[df.columns.str.match("^" + element)].tolist()
                # Remove the product name from the join index
                index = self.index_merge.copy()
                index.remove("faostat_name")
                df = df[index + cols]
                df_all = df_all.merge(df, "left", index)

        # Set an index
        df_all.set_index(self.index, inplace=True)
        return df_all

    def get_country_rows(self, *args, **kwargs):
        """Get only the country rows from the joint_sheets method"""
        df = self.join_sheets(*args, **kwargs)
        selector = df.index.isin(self.country_aggregates, level="country")
        # Copy the slices to new data frames to avoid the warning:
        #   "A value is trying to be set on a copy of a slice from a DataFrame.
        #   Try using .loc[row_indexer,col_indexer] = value instead"
        return df[~selector].copy()

    def get_agg_rows(self, *args, **kwargs):
        """Get only the aggregated rows from the join_sheets method

        Aggregates are world and continents.
        """
        df = self.join_sheets(*args, **kwargs)
        selector = df.index.isin(self.country_aggregates, level="country")
        return df[selector].copy()

    def get_names(self):
        """Get the product and country names from the names sheet"""
        csv_file_name = self.input_dir / "names.csv"
        # Headers are on two columns, load them, merge them and convert to lower case
        df = pandas.read_csv(csv_file_name, header=[0, 1])
        df.columns = [str("_".join(col)).lower() for col in df.columns]
        df.rename(
            columns=lambda x: re.sub(r"unnamed_\d+_", "", str(x)).lower(), inplace=True
        )
        df.rename(columns=lambda x: re.sub(r"_\d+_", "_", str(x)).lower(), inplace=True)
        return df

    def get_country_groups(self):
        """Get the country grouping by continents"""
        df = self.get_names()
        df.rename(
            columns={
                "gfpm_x_country": "country",  # Spreadsheet B2018 and B2020
                "faostat_gfpm_x_country": "country",  # Spreadsheet B2021
                "code": "faostat_country_code",
            },
            inplace=True,
        )
        df["region"] = df["region"].str.upper()
        assert set(["country", "region"]).issubset(df.columns)
        return df

    def convert_sheets_to_dataset(
        self, product: str, other_element: [list, str] = None
    ) -> xarray.Dataset:
        """Combine 2D and 1D arrays from all sheet sheet corresponding to the given
        product into an xarray dataset

        Example load sawnwood data, add GDP to the panel data:

            >>> from cobwood.gfpmx_data import gfpmx_data
            >>> sawn = gfpmx_data.convert_sheets_to_dataset("sawn")
            >>> print(sawn.data_vars)
            >>> panel = gfpmx_data.convert_sheets_to_dataset("panel", ["gdp"])

        """
        sheets = self.sheets[self.sheets["product"] == product]
        # Join other sheets if requested
        if other_element is not None:
            other_sheets = self.sheets[self.sheets["product"] == "other"]
            other_sheets = other_sheets[other_sheets["element"].isin(other_element)]
            sheets = pandas.concat([sheets, other_sheets])
        # Create a dataset for this product
        ds = xarray.Dataset()
        # Add metadata
        ds.attrs["product"] = product
        # Load all sheets for this product
        for this_sheet in sheets.index:
            if this_sheet in ["author", "names", "notes", "worldprice"]:
                continue
            try:
                df = self.get_sheet_wide(this_sheet)
            except pandas.errors.EmptyDataError:
                print(f"The '{this_sheet}' sheet doesn't contain any data.")
                continue
            if "country" not in df.columns:
                print(f"The '{this_sheet}' sheet doesn't have a country column.")
                continue
            element = sheets.loc[this_sheet]["element"]
            # Add the 2D array to the dataset
            ds[element] = convert_to_2d_array(df)
            # Add the 1D arrays to the dataset
            # If `coef_keywords` are in the column names
            # Note: trend and stock elasticity are defined for the World only
            # and could be stored as an attribute, we keep them as a 1D array
            # for now.
            coef_keywords = "elast|const|marginal|rate|stock|trend"
            coefficients = df.columns[df.columns.str.contains(coef_keywords)]
            for col in coefficients:
                # coerce to a numeric value
                df[col] = pandas.to_numeric(df[col], errors="coerce")
                ds[element + "_" + col] = convert_to_1d_array(df, col)
        check_variable_presence(ds)
        # Add region data on continents to be used for groupings
        region_data = self.country_groups.set_index("country")["region"]
        ds["region"] = xarray.DataArray.from_series(region_data)
        # Add country boolean to be used for computations with `loc`
        ds["c"] = ~ds.region.isnull()
        return ds



"""Run the GFPMX model and store output data

"""

from functools import cached_property
from typing import Union, List
import json
import xarray
import cobwood
import pandas

from cobwood.gfpmx_data import GFPMXData
from cobwood.gfpmx_data import compare_to_ref
from cobwood.gfpmx_data import convert_to_2d_array
from cobwood.gfpmx_data import remove_after_base_year_and_copy
from cobwood.gfpmx_equations import compute_one_time_step
from cobwood.gfpmx_plot import facet_plot_by_var
from cobwood.scenario import parse_scenario_yaml


class GFPMX:
    """
    GFPMX model simulation object.

    - Reads data from the GFPMXData object
    - Runs the model
    - Saves the model output in NETCDF files

    Run with xarray and compare to the reference dataset for each available model
    version (with different base years)

        >>> from cobwood.gfpmx import GFPMX
        >>> # Base 2021
        >>> gfpmxb2021 = GFPMX(input_dir="gfpmx_base2021", base_year=2021, scenario="base_2021", rerun=True)
        >>> gfpmxb2021.run_and_compare_to_ref()
        >>> gfpmxb2021.run()

    Load output data, after a run has already been completed

        >>> gfpmx_pikssp2 = GFPMX(input_dir="gfpmx_base2021", base_year=2021, scenario="pikssp2_fel1")

    You can debug data issues by creating the data object only as follows:

        >>> from cobwood.gfpmx_data import GFPMXData
        >>> gfpmx_data_b2018 = GFPMXData(data_dir="gfpmx_8_6_2021", base_year=2018)

    You can debug equations for the different model versions as follows:

        >>> from cobwood.gfpmx_equations import world_price
        >>> world_price(gfpmx_base_2018.sawn, gfpmx_base_2018.indround,2018)

    Run other base years and compare GFPMx Excel results with the one from the cobwood

        >>> # Base 2018
        >>> gfpmxb2018 = GFPMX(input_dir="gfpmx_8_6_2021", base_year=2018, scenario="base_2018")
        >>> # Run and stop when the result diverges from the reference spreadsheet
        >>> gfpmxb2018.run(compare=True)
        >>> # Run and continue when the result diverges (just print the missmatch message)
        >>> gfpmxb2018.run(compare=True, strict=False)
        >>> # Just run, without comparison (default is compare=False)
        >>> gfpmxb2021.run()
        >>> print(gfpmxb2018.indround)
        >>> # Base 2020
        >>> gfpmxb2020 = GFPMX(input_dir="gfpmx_base2020", base_year=2020, scenario="base_2020")
        >>> gfpmxb2020.run_and_compare_to_ref() # Fails
        >>> gfpmxb2021 = GFPMX(input_dir="gfpmx_base2021", base_year=2021, scenario="base_2021")

    You will then be able to load Xarray datasets with the
    `convert_sheets_to_dataset()` method:

        >>> from cobwood.gfpmx_data import GFPMXData
        >>> gfpmxb2018 = GFPMX(input_dir="gfpmx_8_6_2021", base_year=2018)
        >>> print(gfpmxb2018.other_ref)
        >>> print(gfpmxb2018.indround_ref)
        >>> print(gfpmxb2018.sawn_ref)
        >>> print(gfpmxb2018.panel_ref)
        >>> print(gfpmxb2018.pulp_ref)
        >>> print(gfpmxb2018.paper_ref)
        >>> print(gfpmxb2018.gdp)
    """

    def __init__(self, scenario, rerun=False):
        """
        Initialize the GFPMX model with the specified scenario.

        Parameters:
        -----------
        scenario : str
            Name of the scenario to load configuration from
        rerun : bool, optional
            Whether to rerun the model (default: False)
        """
        self.scenario = scenario
        # Parse the YAML scenario configuration file
        self.scenario_yaml_path = cobwood.data_dir / "scenario" / f"{scenario}.yaml"
        self.config = parse_scenario_yaml(self.scenario_yaml_path)
        self.input_dir = cobwood.data_dir / "gfpmx_input" / self.config["input_dir"]
        self.base_year = self.config["base_year"]
        self.output_dir = cobwood.data_dir / "gfpmx_output" / scenario
        self.combined_netcdf_file_path = self.output_dir / "combined_datasets.nc"
        self.last_time_step = 2070
        self.scenario = scenario
        self.products = ["indround", "fuel", "sawn", "panel", "pulp", "paper"]

        # Load reference data
        for product in self.products + ["other"]:
            self[product + "_ref"] = self.input_data.convert_sheets_to_dataset(product)
        self["gdp"] = convert_to_2d_array(self.input_data.get_sheet_wide("gdp"))

        # If the output directory already exists, load data from the netcdf
        # output files, unless explicitly asked to rerun the simulation.
        if self.output_dir.exists() and not rerun:
            print(f"Loading simulation output from netcdf files in {self.output_dir}.")
            self.read_datasets_from_netcdf()
        else:
            # If asked to rerun the first message should not appear
            msg = ""
            if not rerun:
                msg = "There is no output from a previous run for this scenario "
                msg += f"'{self.scenario}'.\n"
            msg += f"Load input data from {self.input_dir} and reset time series to a "
            msg += f"base year {self.base_year} before simulation start."
            print(msg)
            for product in self.products + ["other"]:
                self[product] = remove_after_base_year_and_copy(
                    self[product + "_ref"], self.base_year
                )
        # Create a plot dir
        self.plot_dir = self.output_dir / "plot"
        if not self.plot_dir.exists():
            self.plot_dir.mkdir(parents=True)

    @cached_property
    def input_data(self):
        """Input data"""
        return GFPMXData(self)

    def __getitem__(self, key):
        """Get a dataset from the data dictionary"""
        return getattr(self, key, None)

    def __setitem__(self, key, value):
        """Set a dataset from the data dictionary"""
        setattr(self, key, value)

    def _invalidate_cache(self):
        """Invalidate cached properties

        In case the property has not been used yet, attempting to delete it
        would raise an AttributeError: 'GFPMX' object has no attribute
        'all_products_ds'. That is why the deletion is wrapped in a try except
        statement.

        Show the impact on the all_products ds by changing roundwood production values in 2000

        >>> from cobwood.gfpmx import GFPMX
        >>> model = GFPMX(scenario="base_2021")
        >>> selector_1_prod = dict(country="France", year=2000)
        >>> selector_all_prod = dict(country="France", product="indround", year=2000)
        >>> print(model["indround"]["prod"].loc[selector_1_prod].item())
        >>> print(model.all_products_ds["prod"].loc[selector_all_prod].item())
        >>> print("When we change the product data set. Setting a value to zero.")
        >>> model["indround"]["prod"].loc[selector_1_prod] = 0
        >>> print(model["indround"]["prod"].loc[selector_1_prod].item())
        >>> print("the all_products_ds remains unchanged")
        >>> print(model.all_products_ds["prod"].loc[selector_all_prod].item())
        >>> print("Unless we clear the cache")
        >>> model._invalidate_cache()
        >>> print(model.all_products_ds["prod"].loc[selector_all_prod].item())

        """
        try:
            del self.all_products_ds
        except AttributeError:
            pass

    def run_and_compare_to_ref(self, *args, **kwargs):
        """Takes a gpfmx_data object, remove data after the base year
        run the model and compare it to the reference dataset
        """
        self.run(compare=True, *args, **kwargs)

    def run(self, compare: bool = False, rtol: float = None, strict: bool = True):
        """Run the model for many time steps from base_year + 1 to last_time_step."""
        if rtol is None:
            rtol = 1e-2
        print(f"Running {self.scenario}")
        # Add GDP projections to secondary products datasets.
        # GDP are projected to the future and `self.gdp` might be changed by
        # the user before the model run. This is why it is added only at this time.
        self.sawn["gdp"] = self.gdp
        self.panel["gdp"] = self.gdp
        self.fuel["gdp"] = self.gdp
        self.paper["gdp"] = self.gdp

        for this_year in range(self.base_year + 1, self.last_time_step + 1):
            print(f"Computing: {this_year}", end="\r")
            compute_one_time_step(
                self.indround,
                self.fuel,
                self.pulp,
                self.sawn,
                self.panel,
                self.paper,
                self.other,
                this_year,
            )
            if compare:
                ciepp_vars = ["cons", "imp", "exp", "prod", "price"]
                for product in self.products:
                    compare_to_ref(
                        self[product],
                        self[product + "_ref"],
                        ciepp_vars,
                        this_year,
                        rtol=rtol,
                        strict=strict,
                    )
                compare_to_ref(
                    self.other,
                    self.other_ref,
                    ["stock"],
                    this_year,
                    rtol=rtol,
                    strict=strict,
                )

        # Clear the cache so that results get updated in the combined data set
        self._invalidate_cache()

        # Save simulation output
        self.write_datasets_to_netcdf()

    @cached_property
    def all_products_ds(self):
        """Combined dataset for all products

        The dataset has a third product dimension on top of the country and
        year dimensions
        Examples
        --------
        Extract industrial roundwood production data:

        >>> from cobwood.gfpmx import GFPMX
        >>> gfpmx_base_2021 = GFPMX(scenario="base_2021")
        >>> ds = gfpmx_base_2021.all_products_ds
        >>> print(ds)
        >>> print(ds.individual_dataset_attributes)

        """
        datasets_to_combine = []
        attributes_dict = {}

        for product in self.products:
            # Assign a new coordinate 'product' to each Dataset
            ds = self[product].assign_coords(product=product)
            # Expand dimensions to include 'product'
            ds = ds.expand_dims("product")
            datasets_to_combine.append(ds)
            # Save attributes
            attributes_dict[product] = ds.attrs

        # Concatenate along the new 'product' dimension
        combined_ds = xarray.concat(datasets_to_combine, dim="product")
        # Store attributes as a global attribute in the combined dataset
        attributes_json_str = json.dumps(attributes_dict)
        combined_ds.attrs["individual_dataset_attributes"] = attributes_json_str
        return combined_ds

    def write_datasets_to_netcdf(self):
        """Write all datasets to a single netcdf file with an extra 'product'
        dimension.

        This should be performed after the simulation run, to preserve the
        output of a given scenario. The GFPMX class contains one dataset for
        each product. This method combines them into one combined dataset with
        a new product dimension. It then saves that combined dataset to a
        NetCDF file.
        """
        if not self.output_dir.exists():
            self.output_dir.mkdir(parents=True)

        # Save the products_combined Dataset to NetCDF
        self.all_products_ds.to_netcdf(self.combined_netcdf_file_path)

        # Save other dataset to NETCDF
        self.other.to_netcdf(self.output_dir / "other.nc")

    def read_datasets_from_netcdf(self):
        """Read datasets from a single netcdf file and populate GFPMX object attributes.

        This should be performed to restore the GFPMX object from saved scenarios.
        """
        if not self.combined_netcdf_file_path.exists():
            raise FileNotFoundError(
                f"File {self.combined_netcdf_file_path} does not exist."
            )

        # Read the combined dataset from the NetCDF file
        combined_ds = xarray.open_dataset(self.combined_netcdf_file_path)
        # Retrieve stored attributes and deserialize from JSON string
        attributes_json_str = combined_ds.attrs.get(
            "individual_dataset_attributes", "{}"
        )
        attributes_dict = json.loads(attributes_json_str)

        for product in self.products:
            # Select data corresponding to each product and drop 'product' coordinate
            ds = combined_ds.sel(product=product).drop("product")
            # Restore attributes
            ds.attrs = attributes_dict.get(product, {})
            self[product] = ds

        # Read the other dataset that doesn't have a product dimension
        self["other"] = xarray.open_dataset(self.output_dir / "other.nc")

    def facet_plot_by_var(self, product, *args, **kwargs):
        """Plot one variable for each facet for the given product

        Example use:

            >>> from cobwood.gfpmx import GFPMX
            >>> gfpmx_base2021 = GFPMX(
            ...     input_dir="gfpmx_base2021",
            ...     base_year=2021,
            ...     scenario="base_2021",
            ...     rerun=False
            ... )
            >>> gfpmxb2021.facet_plot("indround")
            >>> # The country argument can specify one line by country
            >>> gfpmxb2021.facet_plot("indround", countries=["Canada", "France", "Japan"])
            >>> # The variable argument can specify one variable by facet
            >>> gfpmxb2021.facet_plot("other", variables=["area", "stock"],
            >>>                  ylabel="Area in 1000ha and stock in million m3")

        """
        accepted_products = self.products + ["other"]
        if product not in accepted_products:
            raise ValueError(f"Product {product} not in {accepted_products}")
        g = facet_plot_by_var(self[product], *args, **kwargs)

    def get_df(
        self, product: Union[str, List[str]], var: Union[str, List[str]]
    ) -> pandas.DataFrame:
        """
        Extract time series data for given product(s) and variable(s) from a GFPMX model.

        Parameters
        ----------
        product : Union[str, List[str]]
            The product identifier(s) (e.g., "indround" for industrial roundwood,
            or ["indround", "fuel"] for multiple products).
        var : Union[str, List[str]]
            The variable identifier(s) (e.g., "prod" for production, or ["prod", "imp"]
            for multiple variables).

        Returns
        -------
        pandas.DataFrame
            DataFrame containing the time series data with reset index and renamed
            variable column(s) using the format "{product}_{var}". The scenario column
            is added as the first column.

        Examples
        --------
        Extract industrial roundwood production data:

        >>> from cobwood.gfpmx import GFPMX
        >>> gfpmx_pikssp2 = GFPMX(scenario="pikssp2")
        >>> irw_prod = gfpmx_pikssp2.get_df(product="indround", var="prod")

        Extract multiple variables for industrial roundwood:

        >>> irw = gfpmx_pikssp2.get_df(product="indround", var=["prod", "imp"])

        Extract single variable for multiple products:

        >>> irw_fw_prod = gfpmx_pikssp2.get_df(product=["indround", "fuel"], var="prod")

        Extract multiple variables for multiple products:

        >>> irw_fw_prod_imp = gfpmx_pikssp2.get_df(
        ...     product=["indround", "fuel"],
        ...     var=["prod", "imp"]
        ... )

        """
        # Convert single values to lists for uniform handling
        if isinstance(product, str):
            product = [product]
        if isinstance(var, str):
            var = [var]

        # Select the specified products from the combined dataset
        df = self.all_products_ds.sel(product=product)[var].to_dataframe().reset_index()

        # Multiple variables: create columns like indround_prod, indround_imp,
        # fuel_prod, fuel_imp
        value_vars = var
        df_melted = df.melt(
            id_vars=["country", "year", "product"],
            value_vars=value_vars,
            var_name="variable",
        )
        df_melted["product_var"] = df_melted["product"] + "_" + df_melted["variable"]
        df = df_melted.pivot_table(
            index=["country", "year"],
            columns="product_var",
            values="value",
            aggfunc="first",
        ).reset_index()
        df.columns.name = None

        # Add scenario column and move it to the front
        df["scenario"] = self.scenario
        cols = df.columns.to_list()
        cols = cols[-1:] + cols[:-1]
        df = df[cols]

        return df
